# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_readers.ipynb.

# %% auto 0
__all__ = ['EXAMPLE_DATA', 'wearable_schema', 'WearableData', 'combine_wearable_streams', 'read_standard_csv',
           'read_standard_json', 'read_actiwatch']

# %% ../nbs/02_readers.ipynb 4
from copy import deepcopy
import datetime
from datetime import datetime
import numpy as np
import pandas as pd
import pylab as plt
from dataclasses import dataclass
from scipy.stats import linregress

import json
import gzip
from typing import List, Tuple, Dict, Union, Optional, Any, Callable, Iterable
import os
from .utils import *
import random
import difflib

import os
from scipy.ndimage import gaussian_filter1d
from os import read
pd.options.mode.chained_assignment = None

import glob
import os
import random
import difflib
from scipy.ndimage import gaussian_filter1d

from  jsonschema import validate
from fastcore.basics import *
from .plots import Actogram
import circadian
from pathlib import Path


# %% ../nbs/02_readers.ipynb 5
# path to the data folder
EXAMPLE_DATA = circadian.__path__[0]

# %% ../nbs/02_readers.ipynb 6
wearable_schema = {
    type: "object",
    "properties": {
            "steps": {
                "type": "array",
                "items": { "type": "object", "properties": { 
                    "start": { "type": "number" },
                    "end": { "type": "number" },
                    "steps": { "type": "number" } } },
                "minItems": 1
            },
            "wake": {
                "type": "array",
                "items": { "type": "object", "properties": { 
                    "start": { "type": "number" },
                    "end": { "type": "number" },
                    "wake": { "type": "number" } } }
            },
            "heartrate": {
                "type": "array",
                "items": { "type": "object", "properties": { 
                    "timestamp": { "type": "number" },
                    "heartrate": { "type": "number" },
            } } },
    },
    "required": ["steps", "wake", "heartrate"]
}
    

# %% ../nbs/02_readers.ipynb 9
@dataclass
class WearableData:
    
    _dataframe: pd.DataFrame # the dataframe that holds the data must have datetime columns plus any other wearable streams from steps, heartrate, wake, light_estimate, activity
    phase_measure: np.ndarray = None
    phase_measure_times: np.ndarray = None
    subject_id: str = "unknown-subject"
    data_id: str = "unknown-data-id"
    meta_data: Dict[str, Any] = None
    
    @property
    def datetime(self) -> np.ndarray:
        return self._dataframe["datetime"].values
    
    @property
    def timestamp(self) -> np.ndarray:
        return self._dataframe["timestamp"].values
    
    @property
    def time_total(self) -> np.ndarray:
        return self._dataframe["time_total"].values
    
    @property
    def light_estimate(self) -> np.ndarray:
        if 'light_estimate' in self._dataframe.columns:
            return self._dataframe["light_estimate"].values
        if 'steps' in self._dataframe.columns:
            return self._dataframe["steps"].values
        if 'activity' in self._dataframe.columns:
            return self._dataframe["activity"].values
        return np.ones_like(self.time_total)*np.nan
    
    @property
    def steps(self) -> np.ndarray:
        if 'steps' in self._dataframe.columns:
            return self._dataframe["steps"].values
        else:
            return np.ones_like(self.time_total)*np.nan
    
    @property
    def activity(self) -> np.ndarray:
        if 'activity' in self._dataframe.columns:
            return self._dataframe["activity"].values
        else:
            return np.ones_like(self.time_total)*np.nan
    
    @property
    def heartrate(self) -> np.ndarray:
        if 'heartrate' in self._dataframe.columns:
            return self._dataframe["heartrate"].values
        else:
            return np.ones_like(self.time_total)*np.nan # generate a nan array
    
    @property
    def wake(self) -> np.ndarray:
        if 'wake' in self._dataframe.columns:
            return self._dataframe["wake"].values
        else:
            return np.ones_like(self.time_total)*np.nan # generate a nan array
        
    @property
    def time_hour_bounds(self) -> Tuple[float, float]:
        return (self.time_total[0], self.time_total[-1])
    
    @property
    def date_bounds(self):
        start_date = pd.to_datetime(self.datetime[0], unit='s')
        end_date = pd.to_datetime(self.datetime[-1], unit='s')
        return (start_date, end_date)
    
    @staticmethod
    def utc_to_hrs(d: datetime):
        return d.hour+d.minute/60.0+d.second/3600.0

    def __post_init__(self):
        # Check that we have the required columns
        assert "datetime" in self._dataframe.columns
        assert "time_total" in self._dataframe.columns
        
    def _copy_with_metadata(self, df: pd.DataFrame) -> "WearableData":
        return WearableData(df, 
                            self.phase_measure, 
                            self.phase_measure_times, 
                            self.subject_id, 
                            self.data_id, 
                            meta_data=self.meta_data)

    def build_sleep_chunks(self, chunk_jump_hrs: float = 12.0) -> List[np.ndarray]:
        time_total = self.time_total
        steps = self.steps
        heartrate = self.heartrate
        wake = self.wake
        data = np.stack((steps, heartrate, wake), axis=0)
        j_idx = np.where(np.diff(time_total) > chunk_jump_hrs)[0]
        return np.split(data, j_idx, axis=1)

    def get_date(self, time_hr: float):
        idx = np.argmin(np.abs(np.array(self.time_total) - time_hr))
        return pd.to_datetime(self.datetime[idx], unit='s')

    def get_timestamp(self, time_hr: float):
        idx = np.argmin(np.abs(np.array(np.hstack(self.time_total)) - time_hr))
        return np.hstack(self.datetime)[idx]

    def trim_by_idx(self, 
                    idx1: int, # First index to keep
                    idx2: int = None # second idx should be greater than idx1, defaults to the last value
                    ) -> 'WearableData':
        df = self._dataframe.loc[idx1:idx2, :]
        return self._copy_with_metadata(df)

    def trim_by_hour(self, 
                     hour_start: float, # First hour to keep
                     hour_end: float, # second hour should be greater than hour_start
                     inplace: bool = False, # if true, the dataframe is modified in place, otherwise a copy is returned
                     ) -> 'WearableData':
        # Trim the __dateframe to be within the interval [t1,t2]
        df = self._dataframe.loc[(self._dataframe.time_total > hour_start) & (self._dataframe.time_total < hour_end)]
        if inplace:
            self._dataframe = df
            return 
        return self._copy_with_metadata(df)
    
    def trim_by_timestamp(self, timestamp_start: float, timestamp_end: float) -> 'WearableData':
        # Trim the __dateframe to be within the interval [t1,t2]
        df = self._dataframe.loc[(self._dataframe.datetime > timestamp_start) & (self._dataframe.datetime < timestamp_end)]
        return self._copy_with_metadata(df)
    
    def __getitem__(self, key: str) -> pd.Series:
        return self._dataframe[key]
        
    def head(self, n: int = 5) -> pd.DataFrame:
        return self._dataframe.head(n)
    
    def tail(self, n: int = 5) -> pd.DataFrame:
        return self._dataframe.tail(n) 
    
    def filter(self, filter_fn: Callable[[pd.DataFrame], pd.DataFrame]) -> 'WearableData':
        return self._copy_with_metadata(filter_fn(self._dataframe))
    
    def aggregate(self, agg_fn: Callable[[pd.DataFrame], pd.DataFrame]) -> 'WearableData':
        return self._copy_with_metadata(agg_fn(self._dataframe))
    
    def groupby(self, by: str) -> 'WearableData':
        return self._copy_with_metadata(self._dataframe.groupby(by))
    
    def join(self, other: 'WearableData', how = 'inner') -> 'WearableData':
        return self._copy_with_metadata(self._dataframe.join(other._dataframe, on='datetime', how=how))
    
    def to_json(self, filename: str = None):
        df_trimed = self._dataframe.drop(columns=['datetime'])
        json_dict = {
            'meta_data' : self.meta_data, 
            'phase_measure' : list(self.phase_measure) if self.phase_measure is not None else [],
            'phase_measure_times' : list(self.phase_measure_times) if self.phase_measure_times is not None else [],
            'subject_id' : self.subject_id,
            'data-id' : self.data_id,
            'wearable' : df_trimed.to_dict()
        }
        if filename is None:
            filename = 'wearable_' + self.subject_id + "_" + self.data_id + '.json'
        with open(filename, 'w', encoding='utf-8') as f:
            json.dump(json_dict, f, ensure_ascii=False,
                        indent=4, cls = NpEncoder )

    @staticmethod
    def from_json(filename) -> 'WearableData':
        jdict = json.load(open(filename, 'r'))
        df = pd.DataFrame.from_dict(jdict['wearable']) 
        df['datetime'] = pd.to_datetime(df['timestamp'], unit='s')
        wdata = WearableData(df)
        for s in jdict.keys():
            if s != 'wearable':
                if isinstance(jdict[s], list):
                    setattr(wdata, s, [np.array(jdict[s])])
                else:
                    setattr(wdata, s, jdict[s])

        return wdata


# %% ../nbs/02_readers.ipynb 10
@patch 
def steps_hr_loglinear(self: WearableData
                       ) -> Tuple[float, float]:
        """
        Find the log steps to hr linear regression parameters .
        hr=beta*log(steps+1.0)+alpha
        Returns beta,alpha
        """
        x = np.log(np.hstack(self.steps)+1.0)
        y = np.hstack(self.heartrate)
        x = x[y > 0]
        y = y[y > 0]
        slope, intercept, r_value, p_value, std_err = linregress(x, y)
        return slope, intercept

# %% ../nbs/02_readers.ipynb 12
@patch
def plot_heartrate(self: WearableData, 
                   t1=None, 
                   t2=None, 
                   ax: plt.Axes = None,
                   show_plot: bool = True,
                   color: str = 'red', 
                   use_dates: bool = True,
                   *args, 
                   **kwargs
                   ) -> plt.Axes:
        t1 = t1 if t1 is not None else self.time_total[0]
        t2 = t2 if t2 is not None else self.time_total[-1]
        wDataTrimmed = self.trim_by_hour(t1, t2)

        hr = deepcopy(wDataTrimmed.heartrate)
        hr[hr == 0] = np.nan
        if ax is None:
            fig = plt.figure()
            ax = plt.gca()
        
        if use_dates:
            x = pd.to_datetime(wDataTrimmed.date_time, unit='s')
        else:
            x = wDataTrimmed.time_total / 24.0 
        ax.plot(x, hr, color=color, *args, **kwargs)
        ax.set_xlabel('Days')
        ax.set_ylabel('BPM')
        ax.set_title('Heart Rate Data')
        
        if show_plot:
            plt.show()
        return ax
      

# %% ../nbs/02_readers.ipynb 13
@patch
def scatter_hr_steps(self: WearableData, 
                     take_log: bool = True, # Log transform the data?
                     *args, 
                     **kwargs):
    
        fig = plt.figure()
        ax = plt.gca()

        steps = np.hstack(self.steps)
        heartrate = np.hstack(self.heartrate)

        if take_log:
            ax.scatter(np.log10(steps[heartrate > 0]+1.0),
                       np.log10(heartrate[heartrate > 0]),
                       color='red',
                       *args,
                       **kwargs)
        else:
            ax.scatter(steps[heartrate > 0], heartrate[heartrate > 0],
                       color='red',
                       *args,
                       **kwargs)

        ax.set_ylabel('BPM')
        ax.set_xlabel('Steps')
        ax.set_title('Heart Rate Data')
        plt.show()

# %% ../nbs/02_readers.ipynb 14
@patch
def plot_hr_steps(self: WearableData, 
                  t1: float = None, 
                  t2: float = None, 
                  *args, 
                  **kwargs):

        time_start = t1 if t1 is not None else self.time_total[0]/24.0
        time_end = t2 if t2 is not None else self.time_total[-1]/24.0

        fig = plt.figure()
        gs = fig.add_gridspec(2, hspace=0.0)
        ax = gs.subplots(sharex=True)
        fig.suptitle(
            f"{self.data_id} Wearable Data: Subject {self.subject_id}")
        hr_all_nan = np.all(np.isnan(self.heartrate))
        if not hr_all_nan:
            ax[0].plot(self.time_total / 24.0, 
                       self.heartrate, 
                       color='red', 
                       *args, 
                       **kwargs)
            
        ax[1].plot(self.time_total / 24.0, 
                   self.steps,
                   color='darkgreen', 
                   *args, 
                   **kwargs)

        sleep_all_nan = np.all(np.isnan(self.wake))
        if not sleep_all_nan:
            ax[1].plot(self.time_total / 24.0, np.array(self.wake) *
                       max(np.median(self.steps), 50.0), color='k')

        if self.phase_measure_times is not None:
            [ax[0].axvline(x=_x / 24.0, ls='--', color='blue')
             for _x in self.phase_measure_times]
            [ax[1].axvline(x=_x / 24.0, ls='--', color='blue')
             for _x in self.phase_measure_times]

        ax[1].set_xlabel("Days")
        ax[0].set_ylabel("BPM")
        ax[1].set_ylabel("Steps")
        ax[0].grid()
        ax[1].grid()
        ax[0].set_xlim((time_start, time_end))
        ax[1].set_xlim((time_start, time_end+3.0))
        ax[0].set_ylim((0, 200))
        plt.show()


# %% ../nbs/02_readers.ipynb 16
def combine_wearable_streams(steps: pd.DataFrame, # dataframe with columns 'start', 'end', 'steps'
                                heartrate: pd.DataFrame, # dataframe with columns 'timestamp', 'heartrate'
                                wake: pd.DataFrame, # dataframe with columns 'start', 'end', 'wake'
                                bin_minutes: int = 6, # bin size in minutes for the resampled combined data
                                subject_id: str="unknown-subject",
                                data_id: str ="Exporter",
                                sleep_trim: bool = False, # drop any entries without a sleep-wake entry
                                inner_join: bool = False # if true, only keep entries that have both heartrate and sleep data
                                ) -> WearableData:
    
    # Convert unix times to datetime
    steps['start'] = pd.to_datetime(steps.start, unit='s')
    steps['end'] = pd.to_datetime(steps.end, unit='s')
    wake['start'] = pd.to_datetime(wake.start, unit='s')
    wake['end'] = pd.to_datetime(wake.end, unit='s')
    heartrate['timestamp'] = pd.to_datetime(heartrate.timestamp, unit='s')
    
    # Resample the steps to the desired bin size
    s1 = steps.loc[:, ['start', 'steps']]
    s2 = steps.loc[:, ['end', 'steps']]
    s1.rename(columns={'start': 'timestamp'}, inplace=True)
    s2.rename(columns={'end': 'timestamp'}, inplace=True)
    steps = pd.concat([s1, s2])
    steps.set_index('timestamp', inplace=True)
    steps = steps.resample(str(int(bin_minutes)) +
                            'Min').agg({'steps': 'sum'})
    steps.reset_index(inplace=True)
    
    # Resample the heartrate data to the desired bin size
    heartrate.set_index('timestamp', inplace=True)
    heartrate = heartrate.resample(
        str(int(bin_minutes))+'Min').agg({'heartrate': 'max'})
    heartrate.reset_index(inplace=True)

    # Merge the steps and heartrate data and fill missing heartrate with zeros
    merge_method = 'inner' if inner_join else 'left'
    df = pd.merge(steps, heartrate, on='timestamp', how=merge_method)
    
    # Resample the wake data to the desired bin size
    s1 = wake.loc[:, ['start', 'wake']]
    s2 = wake.loc[:, ['end', 'wake']]
    s1.rename(columns={'start': 'timestamp'}, inplace=True)
    s2.rename(columns={'end': 'timestamp'}, inplace=True)
    wake = pd.concat([s1, s2])
    wake.set_index('timestamp', inplace=True)
    wake = wake.resample(str(int(bin_minutes)) +
                            'Min').agg({'wake': 'max'})
    wake.reset_index(inplace=True)
    
    merge_method = 'inner' if inner_join else 'left'
    df = pd.merge(df, wake, on='timestamp', how=merge_method)
    
    df['datetime'] = df['timestamp'] 
    
    # Make the timestamp column actually be a unix timestamp
    df['timestamp'] = (
        df['datetime'] - pd.Timestamp("1970-01-01")) // pd.Timedelta('1s')
    
    time_start = WearableData.utc_to_hrs(df.datetime.iloc[0])
    df['time_total'] = time_start + (df.timestamp-df.timestamp.iloc[0])/3600.0
    
    if sleep_trim:
        df.dropna(subset=['wake'], inplace=True)
    
    aw = WearableData(_dataframe=df,
                    subject_id=subject_id,
                    data_id=data_id
                    )

    return aw
    

def read_standard_csv(directory_path: str, # path to the directory containing the csv files
                        bin_minutes: int = 6,
                        subject_id="unknown-subject",
                        data_id="Exporter",
                        sleep_trim: bool = False, # drop any entries without a sleep-wake entry
                        inner_join: bool = False # if true, only keep entries that have both heartrate and sleep data
                        ) -> WearableData:

    steps = pd.read_csv(directory_path+"/steps.csv", columns=['start', 'end', 'steps'])
    heartrate = pd.read_csv(directory_path+"/heartrate.csv", columns=['timestamp', 'heartrate'])
    wake = pd.read_csv(directory_path+"/sleep.csv", columns=['start', 'end', 'wake'])
    
    return combine_wearable_streams(steps, heartrate, wake, bin_minutes, subject_id, data_id, sleep_trim, inner_join)

        

def read_standard_json( filepath: str, # path to json file
                        bin_minutes: int = 6, # data will be binned to this resolution in minutes
                        subject_id: str ="unknown-subject", #subject id to be used
                        data_id: str = "Exporter", # name of the data source
                        gzip_opt: bool = False, # set to true if the file is gzipped, will be autodetected if extension is .gz
                        sleep_trim: bool = False, # drop any entries without a sleep-wake entry
                        inner_join: bool = False # if true, only keep entries that have both heartrate and sleep data
                        ) -> WearableData:
    gzip_opt = gzip_opt if gzip_opt else filepath.endswith(".gz")
    fileobj = gzip.open(filepath, 'r') if gzip_opt else open(filepath, 'r')
    rawJson = json.load(fileobj)
    validate(rawJson, wearable_schema)
    
    steps = pd.DataFrame(rawJson['steps'], columns=["start", "end", "steps"])
    # These could be empty
    wake = pd.DataFrame(rawJson['wake'], columns=["start", "end", "wake"]) 
    heartrate = pd.DataFrame(rawJson['heartrate'], columns=["timestamp", "heartrate"])
    
    return combine_wearable_streams(steps, heartrate, wake, bin_minutes, subject_id, data_id, sleep_trim, inner_join)
        
        


# %% ../nbs/02_readers.ipynb 17
@patch 
def fillna(self: WearableData, 
             column_name: str = "heartrate", # column to fill in the dataframe
             with_value: float = 0.0, # value to fill with
             inplace: bool = False # if true, the WearableData object will be modified in place
             ) -> WearableData:
    
    if inplace:
        self._dataframe[column_name].fillna(with_value, inplace=True)
    else:
        df = self._dataframe.copy()
        filled_column = df[column_name].fillna(with_value)
        df[column_name] = filled_column 
        return self._copy_with_metadata(df)

# %% ../nbs/02_readers.ipynb 33
@patch 
def plot_light_activity(self: WearableData, 
                        show=True, 
                        vlines=None, 
                        *args, **kwargs):

        fig = plt.figure()
        gs = fig.add_gridspec(2, hspace=0)
        ax = gs.subplots(sharex=True)
        fig.suptitle(
            f"{self.data_id} Subject {self.subject_id}")
        ax[0].plot(self.time_total / 24.0, np.log10(self.light_estimate+1.0), color='red')
        ax[1].plot(self.time_total / 24.0, self.activity, color='darkgreen')
        
        try:
            ax[1].plot(self.time_total / 24.0, self.wake *
                       np.median(self.steps), color='k')
        except:
            print(f"Error with wake plot with {self.subject_id}")

        if self.phase_measure_times is not None:
            [ax[0].axvline(x=_x / 24.0, ls='--', color='blue')
             for _x in self.phase_measure_times]
            [ax[1].axvline(x=_x / 24.0, ls='--', color='blue')
             for _x in self.phase_measure_times]

        if vlines is not None:
            [ax[0].axvline(x=_x / 24.0, ls='--', color='cyan')
             for _x in vlines]
            [ax[1].axvline(x=_x / 24.0, ls='--', color='cyan')
             for _x in vlines]

        ax[1].set_xlabel("Days")
        ax[0].set_ylabel("Lux (log 10)")
        ax[1].set_ylabel("Activity Counts")
        ax[0].grid()
        ax[1].grid()
        if show:
            plt.show()
        else:
            return ax


# %% ../nbs/02_readers.ipynb 35
def read_actiwatch(filepath: str, # path to actiwatch csv file
                        MIN_LIGHT_THRESHOLD=5000, # used to trim off empty data at the beginning and end of the file, must reach this amount of light to be included. Turn this off can setting this to 0 or negative
                        round_data=True, # round the data to the nearest bin_minutes
                        bin_minutes=6, # bin the data to this resolution in minutes, only used if round_data is true
                        dt_format: str = None, # format of the date time string, if None, will be inferred
                        data_id: str = "Actiwatch", # name of the data source
                        subject_id: str = "unknown-subject", #subject id to be used
                        ) -> WearableData:
    
    df = pd.read_csv(filepath, names=['Date', 'Time', 'Activity', 'White Light', 'Sleep/Wake'], header=0) 
    df['datetime'] = df['Date']+" "+df['Time']
    if dt_format is None:
        df['datetime'] = pd.to_datetime(df.datetime, infer_datetime_format=True)
    else:
        df['datetime'] = pd.to_datetime(df.datetime, format=dt_format)

    df['UnixTime'] = (
        df['datetime'] - pd.Timestamp("1970-01-01")) // pd.Timedelta('1s')
    
    df.rename(columns={'White Light': 'light_estimate'}, inplace=True)
    df.rename(columns={'Sleep/Wake': 'wake'}, inplace=True)
    df.rename(columns={'Activity': 'activity'}, inplace=True)

    df['light_estimate'].fillna(0, inplace=True)
    df['LightSum'] = np.cumsum(df.light_estimate.values)
    df['LightSumReverse'] = np.sum(
        df.light_estimate.values) - np.cumsum(df.light_estimate.values) + 1.0

    df = df[(df.LightSum > MIN_LIGHT_THRESHOLD) & (
        df.LightSumReverse > MIN_LIGHT_THRESHOLD)]

    time_start = WearableData.utc_to_hrs(df.datetime.iloc[0])
    df2 = df[['UnixTime']].copy(deep=True)
    base_unix_time = df2['UnixTime'].iloc[0]
    df['time_total'] = time_start + \
        (df2.loc[:, ['UnixTime']]-base_unix_time)/3600.0
        
    df['timestamp'] = df['UnixTime']
    
    df = df[["datetime", "timestamp", "time_total", "activity", "light_estimate", "wake"]]
    if round_data:
        df.set_index('datetime', inplace=True)
        df = df.resample(str(int(bin_minutes))+'Min').agg({'time_total': 'min',
                                                            'activity': 'sum',
                                                            'light_estimate': 'median',
                                                            'wake': 'max'})
        df.reset_index(inplace=True)
    df['time_total'].interpolate(inplace=True)
    df.activity.fillna(0.0, inplace=True) 
    df.light_estimate.fillna(0.0, inplace=True)
    return WearableData(
        _dataframe = df,
        data_id=data_id,
        subject_id=subject_id
    )




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_utils.ipynb.

# %% auto 0
__all__ = ['NpEncoder', 'simple_norm_stepshr_sleep_classifier', 'phase_ic_guess', 'abs_hour_diff', 'cut_phases_12',
           'convert_binary', 'cal_days_diff', 'interpolateLinear', 'interpolateLinearExt', 'parse_dt', 'circular_mean',
           'phase_coherence', 'phase_coherence_clock', 'angle_difference', 'subtract_clock_times', 'circular_av_clock',
           'circular_scatter', 'times_to_angle', 'timezone_mapper', 'split_missing_data', 'split_drop_data']

# %% ../nbs/07_utils.ipynb 3
import numpy as np
import pandas as pd
import scipy as sp
from numba import jit
from scipy import interpolate
import torch 
import json
import pytz
import datetime
import copy
import scipy as sp
from scipy.integrate import solve_ivp
from math import *
import pylab as plt


# %% ../nbs/07_utils.ipynb 4
class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return super(NpEncoder, self).default(obj)

# %% ../nbs/07_utils.ipynb 5
pd.options.mode.chained_assignment = None  # default='warn'

def simple_norm_stepshr_sleep_classifier(t):
        t[0,:] = torch.tanh((t[0,:] - 10.0)/ 100.0)
        t[1, torch.nonzero(t[1,:])] = torch.tanh((t[1,torch.nonzero(t[1,:])] - 60.0) / 30.0)
        return t 

def phase_ic_guess(time_of_day: float):
    time_of_day = np.fmod(time_of_day, 24.0)

    # Wake at 8 am after 8 hours of sleep
    # State at 00:00
    psi = 1.65238233

    # Convert to radians, add to phase
    psi += time_of_day * np.pi / 12
    return psi

def abs_hour_diff(x, y):
    """
    function abs_hour_diff(x,y)

    Find the difference in hours between
    two clock times (wrapped)
    """
    a1 = min(x, y)
    a2 = max(x, y)
    s1 = a2-a1
    s2 = 24.0+a1-a2
    return(min(s1, s2))


def cut_phases_12(p):
    """
    Function to make the branch cut for the DLMO times be at 12 instead of 24.
    This is better because lots of DLMOs are near midnight, but many fewer are near
    noon.

        cut_phases_12(timept)
    """

    while (p < 0.0):
        p += 24.0

    p = np.fmod(p, 24.0)

    if p > 12:
        return p-24.0
    else:
        return p


def convert_binary(x, breakpoint: float = 0.50):
    x[x <= breakpoint] = 0.0
    x[x > breakpoint] = 1.0
    return x


def cal_days_diff(a, b):
    """Get the calander days between two time dates"""
    A = a.replace(hour=0, minute=0, second=0, microsecond=0)
    B = b.replace(hour=0, minute=0, second=0, microsecond=0)
    return (A - B).days


@jit(nopython=True)
def interpolateLinear(t, xvals, yvals):
    """Implement a faster method to get linear interprolations of the light functions"""

    if (t >= xvals[-1]):
        return (0.0)
    if (t <= xvals[0]):
        t += 24.0

    i = np.searchsorted(xvals, t) - 1
    ans = (yvals[i + 1] - yvals[i]) / \
          ((xvals[i + 1] - xvals[i]) * (t - xvals[i])) + yvals[i]
    return (ans)


@jit(nopython=True)
def interpolateLinearExt(t, xvals, yvals):
    """Implement a faster method to get linear interprolations of the light functions, exclude non-full days"""
    i = np.searchsorted(xvals, t) - 1
    ans = (yvals[i + 1] - yvals[i]) / \
          ((xvals[i + 1] - xvals[i]) * (t - xvals[i])) + yvals[i]
    return (ans)


def parse_dt(date, time):
    strDate = date + ' ' + time
    return pd.to_datetime(strDate, format='%m/%d/%Y %I:%M %p')


# %% ../nbs/07_utils.ipynb 7
def circular_mean(series):
    Z=complex(0,0)
    series=np.array(series)
    for i in range(len(series)):
        Z+=np.exp(series[i]*complex(0,1))

    Z=Z/float(len(series))

    ans=np.angle(Z)
    if (ans<0.0):
        ans+=2*sp.pi
    return(ans)

def phase_coherence(series):
    Z=complex(0,0)
    series=np.array(series)
    for i in range(len(series)):
        Z+=np.exp(series[i]*complex(0,1))

    Z=Z/float(len(series))

    ans=np.absolute(Z)
    return(ans)

def phase_coherence_clock(series):
    angles=np.pi/12.0*series
    return(phase_coherence(angles))

def angle_difference(c1, c2) -> float:
    """Find the angle between two angles given in radians
    angle_difference(c1, c2)
    c1-c2
    """

    return(np.angle(np.exp(complex(0,1)*(c1-c2))))


def subtract_clock_times(c1, c2):
    """Find the hour differences between two clock times new"""
    a1=sp.pi/12.0*c1
    a2=sp.pi/12.0*c2
    adiff=angle_difference(a1, a2)
    return(12.0/sp.pi*adiff)


def circular_av_clock(series):
    """Find the average time given a list of clock times"""
    angles=sp.pi/12.0*series
    ans_angle=circular_mean(angles)
    #back to clock time
    return(ans_angle*12.0/sp.pi)
    

def circular_scatter(ax, angles, clock_times=False, radius=1.0, color='blue'):
    """Adds a polar scatter plot of clock times to an axes with polar axis i.e.
        ax = plt.subplot(111, polar=True)
        Will also plot the circular mean angle and the phase coherence
    """
    
    angles=np.array(angles)
    radii=radius*np.ones(len(angles))


    if clock_times:
        angles=angles*sp.pi/12.0

    ax.scatter(angles, radii, color=color)
    ax.set_theta_zero_location("N")
    ax.set_theta_direction(-1)
    ax.set_thetagrids(list(range(0,360,45)), list(range(0,24,3)))
    ax.set_rmax(1.2)
    ax.set_rticks([0.0,0.2,0.6,0.8,1.0])
    ax.annotate("", xytext=(0.0,0.0), xy=(circular_mean(angles),phase_coherence(angles)),arrowprops=dict(facecolor=color))
    
    
    


# %% ../nbs/07_utils.ipynb 9
class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)


def times_to_angle(time_vector: np.ndarray):
    """
        Take an array of times and return R, psi 
        giving the mean angle (psi) and amplitude (R)
    """
    rad_vector = np.fmod(time_vector, 24.0) * np.pi/12.0
    Z = np.sum(np.exp(rad_vector*1j))/len(rad_vector)
    return(np.abs(Z), np.angle(Z))


def timezone_mapper(dt_object: datetime, timezone: str = 'America/Detroit'):
    """
        Take in local time as datetime object and give back UTC with 
        day lights savings accounted for as a timestamp
    """

    local_timezone = pytz.timezone(timezone)
    return local_timezone.localize(dt_object).timestamp()


def split_missing_data(date_time, ts, y, hr=None, break_threshold=96.0):

    # Find idx at start and end of long periods of zeros

    idx_start = None
    idx_end = None
    in_region = False
    crop_regions = []

    for (k, t) in enumerate(ts):
        if y[k] <= 0.0 and not in_region:
            idx_start = k
            in_region = True
        if y[k] > 0.0 and in_region:
            idx_end = k-1
            in_region = False
            if ts[idx_end]-ts[idx_start] >= break_threshold:
                crop_regions += [idx_start, idx_end]
    ts_split = np.split(ts, crop_regions)
    y_split = np.split(y, crop_regions)

    if hr is not None:
        hr_split = np.split(hr, crop_regions)

    print(f"Splitting data into {len(y_split)} regions")

    if hr is not None:
        return np.split(date_time, crop_regions), ts_split, y_split, hr_split
    else:
        return np.split(date_time, crop_regions), ts_split, y_split


def split_drop_data(date_time, ts, steps, hr, wake, break_threshold=96.0, min_length: float = 30.0):
    """
        Used to split long JSON into contin data steaks of at 
        least X=30 days.

        Uses that missing data will be zeros for steps and hr  and 
        0.5 for the wake data. 

        min_length is in days
    """

    idx_start = None
    idx_end = None
    in_region = False
    crop_regions = []

    for (k, t) in enumerate(ts):
        if (steps[k] <= 0.0 or hr[k] <= 0 or wake[k] == 0.50) and not in_region:
            idx_start = k
            in_region = True
        if steps[k] > 0.0 and hr[k] > 0 and wake[k] != 0.50 and in_region:
            idx_end = k-1
            in_region = False
            if ts[idx_end]-ts[idx_start] >= break_threshold:
                crop_regions += [idx_start, idx_end]

    ts_split = np.split(ts, crop_regions)
    steps_split = np.split(steps, crop_regions)
    hr_split = np.split(hr, crop_regions)
    wake_split = np.split(wake, crop_regions)
    date_time = np.split(date_time, crop_regions)

    # Find idxs for regions which are longer than min_length

    idx_long = [k for (k, val) in enumerate(ts_split)
                if (val[-1]-val[0])/24.0 >= min_length]

    if len(idx_long) > 0:
        return ([date_time[i] for i in idx_long], [ts_split[idx] for idx in idx_long], [steps_split[i] for i in idx_long],
                [hr_split[i] for i in idx_long], [wake_split[i] for i in idx_long])
    else:
        return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/08_cli.ipynb.

# %% auto 0
__all__ = ['mytz', 'main_acto', 'main_esri']

# %% ../nbs/08_cli.ipynb 4
import torch
from torch import jit
from datetime import datetime
import argparse
from .plots import Actogram
from .models import Hannay19
from .readers import WearableData, read_standard_csv, read_standard_json
from .utils import phase_ic_guess
from .sleep import cluster_sleep_periods_scipy, sleep_midpoint
from .utils import simple_norm_stepshr_sleep_classifier
from datetime import datetime
import matplotlib.pylab as plt
import numpy as np
from pytz import timezone
mytz = timezone('EST')


# %% ../nbs/08_cli.ipynb 5
def main_acto():
    parser = argparse.ArgumentParser(description="""Make an actogram""")

    parser.add_argument('-j', '--json',
                        required=False,
                        action='store',
                        type=str,
                        help="Specify a json file with the data"
                        )
    parser.add_argument('-c', '--csv',
                        required=False,
                        action='store',
                        type=str,
                        help="Specify a directory with csv data in it"
                        )
    parser.add_argument("-sm", "--sleepmodel",
                        required=False,
                        action='store',
                        type=str,
                        help="Machine learning pytorch model for steps and heartrate to wake prediction"
                        )
    parser.add_argument('-d', '--dlmo',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Integrate the model and plot dlmo times"
                        )
    parser.add_argument('-cbt', '--cbt',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Integrate the model and plot core body temperature times"
                        )
    parser.add_argument('--sleep',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Add sleep midpoints"
                        )
    parser.add_argument('-t', '--threshold',
                        required=False,
                        action='store',
                        type=float,
                        help="Threshold for displaying as light on the actogram",
                        default=1.0
                        )
    parser.add_argument('-m', '--multiplier',
                        required=False,
                        action='store',
                        type=float,
                        help="Steps to light multiplier",
                        default=1.0
                        )
    parser.add_argument('-p', '--period',
                        required=False,
                        action='store',
                        default=23.84,
                        type=float,
                        help="Set the SPM period"
                        )
    parser.add_argument('-r', '--raw',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Plot the raw steps and heartrate data"
                        )
    parser.add_argument('--scatter',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Plot the raw steps and heartrate data"
                        )
    parser.add_argument('-t1', '--t1',
                        required=False,
                        type=float,
                        action='store',
                        default=None,
                        help="Trim data before this time, in days"
                        )
    parser.add_argument('-t2', '--t2',
                        required=False,
                        type=float,
                        action='store',
                        default=None,
                        help="Trim data after this time, in days"
                        )
    parser.add_argument('-s', '--sigma',
                        required=False,
                        type=float,
                        action='store',
                        default=0.5,
                        help="Smooth the light(steps) data"
                        )


    args = parser.parse_args()


    def generated_sleep(ml_model_path, steps, hr):
        ml_model = jit.load(ml_model_path)
        data = torch.vstack((torch.tensor(steps), torch.tensor(hr)))
        data = simple_norm_stepshr_sleep_classifier(data).unsqueeze(0).float()
        wake_predicted = torch.sigmoid(ml_model(data)).squeeze(
            0).squeeze(0).detach().numpy()
        len_out = wake_predicted.shape[0]
        wake_predicted = np.hstack((wake_predicted, len(steps) - len_out))
        wake_predicted = np.where(wake_predicted > 0.50, 1.0, 0.0)
        return wake_predicted
    
    if args.json:
        awObj = read_standard_json(args.json)

    if args.csv:
        awObj = read_standard_csv(args.csv)

    if args.csv or args.json:
        if args.t1 or args.t2:
            t1 = args.t1 or 0
            t2 = args.t2 or np.floor(np.squeeze(awObj.time_total)[-1]/24.0)+1
            awObj.trim_data(t1*24.0, t2*24.0)
        hr = awObj.heartrate
        ts = awObj.time_total
        steps = awObj.steps

        if args.sleepmodel:
            awObj.wake = generated_sleep(
                ml_model_path=args.sleepmodel, 
                steps=steps, 
                hr=hr)

    if args.raw:
        awObj.plot()

    ic = np.array([0.70, phase_ic_guess(ts[0]), 0.0])

    if args.scatter:
        awObj.scatter_hr_steps()


    plt.figure()
    ax = plt.gca()
    acto = Actogram(np.hstack(ts),
                    np.hstack(steps),
                    ax=ax,
                    threshold=args.threshold,
                    opacity=1.0,
                    sigma=[args.sigma, args.sigma])

    if awObj.wake is not None and args.sleep:
        print("Adding sleep layer")
        acto = Actogram(np.hstack(ts),
                        np.hstack(awObj.wake),
                        ax=ax,
                        threshold=0.5,
                        smooth=False,
                        opacity=0.50,
                        color='green')

    if args.dlmo:
        ic = np.array([0.70, phase_ic_guess(ts[0]), 0.0])
        spm2 = Hannay19({'tau': args.period})
        dlmo_runs = spm2.integrate_observer(ts, args.multiplier*steps, ic)
        acto.plot_phasemarker(
            dlmo_runs, error=np.ones(len(dlmo_runs)), color='blue')
        print(f"DLMO mean: {np.mean(np.fmod(dlmo_runs, 24.0))}")
        print("Last 14 days of DLMOs")
        print(np.fmod(dlmo_runs[-14:], 24.0))

    if args.cbt:
        ic = np.array([0.70, phase_ic_guess(ts[0]), 0.0])
        spm2 = Hannay19({'tau': args.period})
        cbt_runs = spm2.integrate_observer(
            ts, args.multiplier*steps, ic, observer=SinglePopModel.CBTObs)
        acto.plot_phasemarker(
            cbt_runs, error=np.ones(len(cbt_runs)), color='red')
        print(f"CBT mean: {np.mean(np.fmod(cbt_runs, 24.0))}")
        print("Last 14 days of CBTs")
        print(np.fmod(cbt_runs[-14:], 24.0))


    # if args.sleep:
    #     ts_flat = np.hstack(ts)
    #     steps_flat = np.hstack(steps)
    #     if awObj.wake is None:
    #         spm2 = Hannay19({'tau' : args.period})
    #         sol= spm2.integrate_model(ts_flat, args.multiplier*steps_flat, phase_ic_guess(ts_flat[0]))
    #         wake_score = np.diff(sol[3,:], prepend=0) > 0
    #     else:
    #         print("Using the provided wake column")
    #         wake_score=np.hstack(awObj.wake)

    #     sleep_clusters = cluster_sleep_periods_scipy(wake_score,
    #                                            90.0,
    #                                            makeplot=False)
    #     sleep_mid, duration = sleep_midpoint(ts_flat, sleep_clusters)
    #     acto.plot_phasemarker(sleep_mid, error=(
    #         duration/2.0), color="green", alpha=1.0)
    #     print(f"The median sleep duration is {np.median(duration)}")
    #     print(f"The average bedtime is {np.mean(np.fmod(sleep_mid-0.5*duration, 24.0))}")
    #     print(f"The average waketime is {np.mean(np.fmod(sleep_mid,24.0)+0.5*duration)}")


    plt.show()


# %% ../nbs/08_cli.ipynb 6
import numpy as np
import matplotlib.pyplot as plt
import argparse
from pathlib import Path
import sys
import torch
from torch import jit
import circadian
from .readers import WearableData, read_standard_csv, read_standard_json
from .models import Hannay19
from .utils import phase_ic_guess
from .metrics import *
from .utils import simple_norm_stepshr_sleep_classifier
import pandas as pd
from .plots import Actogram
from pytz import timezone


# %% ../nbs/08_cli.ipynb 7
def main_esri():
    mytz = timezone('EST')
    parser = argparse.ArgumentParser(
        description="""Compute the Entrainment Signal Regularity Index for a data set""")

    parser.add_argument("-a", "--actogram",
                        required=False,
                        action='store_true',
                        default=False
                        )

    parser.add_argument('-j', '--json',
                        required=False,
                        action='store',
                        type=str,
                        help="Specify a json file with the data"
                        )
    parser.add_argument('-c', '--csv',
                        required=False,
                        action='store',
                        type=str,
                        help="Specify a directory with csv data in it"
                        )

    parser.add_argument('-s', '--sigma',
                        required=False,
                        type=float,
                        action='store',
                        default=0.5,
                        help="Smooth the light(steps) data"
                        )

    parser.add_argument('-t1', '--t1',
                        required=False,
                        type=float,
                        action='store',
                        default=None,
                        help="Trim data before this time, in days"
                        )
    parser.add_argument('-t2', '--t2',
                        required=False,
                        type=float,
                        action='store',
                        default=None,
                        help="Trim data after this time, in days"
                        )

    parser.add_argument("-sm", "--sleepmodel",
                        required=False,
                        action='store',
                        type=str,
                        help="Machine learning pytorch model for steps and heartrate to wake prediction"
                        )

    parser.add_argument('-t', '--threshold',
                        required=False,
                        action='store',
                        type=float,
                        help="Threshold for displaying as light on the actogram",
                        default=1.0
                        )

    parser.add_argument('-o', '--outfile',
                        required=False,
                        action='store',
                        type=str,
                        help="Path to the outfile",
                        default="esri_results.csv"
                        )
    parser.add_argument('-np', '--noplot',
                        required=False,
                        action='store_true',
                        help="Use this to suppress the plot",
                        default=False
                        )


    args = parser.parse_args()

    def generated_sleep(ml_model_path, steps, hr):
        ml_model = jit.load(ml_model_path)
        data = torch.vstack((torch.tensor(steps), torch.tensor(hr)))
        data = simple_norm_stepshr_sleep_classifier(data).unsqueeze(0).float()
        wake_predicted = torch.sigmoid(ml_model(data)).squeeze(
            0).squeeze(0).detach().numpy()
        len_out = wake_predicted.shape[0]
        wake_predicted = np.hstack((wake_predicted, len(steps) - len_out))
        wake_predicted = np.where(wake_predicted > 0.50, 1.0, 0.0)
        return wake_predicted


    if args.json:
        awObj = read_standard_json(args.json)

    if args.csv:
        awObj = read_standard_csv(args.csv)

    if args.csv or args.json:
        if args.t1 or args.t2:
            t1 = args.t1 or 0
            t2 = args.t2 or np.floor(np.squeeze(awObj.time_total)[-1]/24.0)+1
            awObj.trim_data(t1*24.0, t2*24.0)
        hr = awObj.heartrate
        ts = awObj.time_total
        steps = awObj.steps

        if args.sleepmodel:
            awObj.wake = generated_sleep(
                ml_model_path=args.sleepmodel, steps=steps, hr=hr)


    def moving_average(x, w):
        return np.convolve(x, np.ones(w), 'valid') / w


    smooth_length = 100
    num_days_compactness = 4.5

    ts_compact, timestamps, compactness = esri_trajectory(
        awObj, gamma=0.0, multiplier=1.0, num_days=num_days_compactness)
    compactness_smooth = moving_average(compactness, smooth_length)
    plt.figure()
    ax1 = plt.gca()
    ax1.plot(np.array(compactness)*24.0, np.array(ts_compact) /
            24.0 + num_days_compactness,  alpha=0.50)
    ax1.plot(compactness_smooth*24.0,
            np.array(ts_compact[smooth_length-1:])/24.0 + num_days_compactness, color='red')
    ax1.set_title(f"ESRI")
    ax1.set_ylabel("Days")
    ax1.set_xlabel("ESRI Score")
    ax1.set_ylim((0, max(np.array(ts_compact) / 24.0)))
    ax1.invert_yaxis()



    print(f"Writing results to {args.outfile}")
    df = pd.DataFrame({'timestamp': timestamps, 'esri': compactness})
    df.to_csv(args.outfile, index=False)

    if not args.noplot:
        acto = Actogram(np.hstack(ts),
                        np.hstack(steps),
                        ax=ax1,
                        threshold=args.threshold,
                        opacity=1.0,
                        sigma=[args.sigma, args.sigma])


        plt.show()




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_models.ipynb.

# %% auto 0
__all__ = ['DynamicalTrajectory', 'CircadianModel', 'Forger99Model', 'Hannay19TP', 'Hannay19']

# %% ../nbs/00_models.ipynb 3
import scipy as sp
from abc import ABC, abstractmethod
import numpy as np
from scipy.signal import find_peaks
from .utils import phase_ic_guess

from ctypes import c_void_p, c_double, c_int, cdll
import scipy as sp
import pylab as plt
from pathlib import Path
import sys
from numba import jit, njit, prange
from typing import List, Tuple, Dict, Union, Optional, Callable



# %% ../nbs/00_models.ipynb 7
class DynamicalTrajectory:
    
    def __init__(self, 
                 ts: np.ndarray, # time points
                 states: np.ndarray # states at time points
                 ) -> None:
        self.ts = ts
        self.states = states
        
    def __call__(self, t: float) -> np.ndarray:
        return np.interp(t, self.ts, self.states)
    
    def __getitem__(self, idx: int) -> Tuple[float, np.ndarray]:
        return self.ts[idx], self.states[idx]
    
    def __len__(self) -> int:
        return len(self.ts)
    
    def get_batch(self,index: int) -> 'DynamicalTrajectory':
        if self.states.ndim >= 3:
            return DynamicalTrajectory(self.ts, self.states[index, :, :])
        else:
            return DynamicalTrajectory(self.ts, self.states)
        
    @property
    def batch_size(self) -> int:
        if self.states.ndim >= 3:
            return self.states.shape[0]
        else:
            return 1
    
    

# %% ../nbs/00_models.ipynb 8
class CircadianModel(ABC):
    """ Abstract base class for circadian models, defines the interface for all models """

    def __init__(self, 
                 params: dict = None # dictionary of parameters for the model, if None then the default parameters are used
                 ):
        """ Creates a new instance of the model """
        pass

    @abstractmethod
    def _default_params(self) -> dict: # dictionary of default parameters for the model:
        """
            Defines the default parameters for the model
        """
        pass
    
    @property
    def get_parameters(self)-> dict:
        """
            Returns the parameters for the model
        """
        pass

    def step_rk4(self,
                 state: np.ndarray, #dy/dt = f(y, t)
                 light_val: float, #light value at time t in lux
                 dt=0.10 #step size in hours 
                 ):
        """
            Return the state of the model assuming a constant light value
            for one time step and using a fourth order Runga-Kutta integrator to perform the step
        """
        k1 = self.derv(state, light=light_val)
        k2 = self.derv(state + k1 * dt / 2.0, light=light_val)
        k3 = self.derv(state + k2 * dt / 2.0, light=light_val)
        k4 = self.derv(state + k3 * dt, light=light_val)
        state = state + (dt / 6.0) * (k1 + 2.0*k2 + 2.0*k3 + k4)
        return state

    def integrate_model(self,
                        ts: np.ndarray,  # Array of time points, also determines step size of RK4 solver
                        light_est: np.ndarray,  # Array of light estimates, should be the same length as ts
                        state: np.ndarray,  # Initial state of the model
                        ) -> DynamicalTrajectory:
        n = len(ts)
        sol = np.zeros((*state.shape, n))
        sol[..., 0] = state
        for idx in range(1, n):
            state = self.step_rk4(
                state=state,
                light_val=light_est[idx],
                dt=ts[idx]-ts[idx-1]
            )
            sol[..., idx] = state
        return DynamicalTrajectory(ts, sol)

    def dlmos(self, 
              trajectory: DynamicalTrajectory,  # solution from integrate_model
              ) -> np.ndarray: #array of times when the dlmo occurs for the model 
        """
            Finds the Dim Light Melatonin Onset (DLMO) markers for the model along a trajectory
        """
        pass

    def cbt(self, 
            trajectory: DynamicalTrajectory,  # solution from integrate_model
            ) -> np.ndarray: # array of times when the cbt occurs
        """
            Finds the core body temperature minumum markers for the model along a trajectory
        """
        pass

    def observer(self,
                 trajectory: np.ndarray,  # solution from integrate_model
                 observer_func: callable, # function that takes the state of the model and returns a scalar will be triggered when the sign of the function changes
                 ) -> np.array: # this will return the times when the observer func changes signs
        """ 
            Defines a generic observer for the model, this will return the times when the observer_func changes sign
        """ 
        pass

    @staticmethod
    def amplitude(state: np.ndarray, #dynamic state of the model
                  ) -> float:
        """
            Gives the amplitude of the model at a given state
        """
        pass

    @staticmethod
    def phase(state: np.ndarray #dynamic state of the model
              ) -> float:
        """
            Gives the phase of the model at a given state
        """
        pass

    @property
    def default_initial_conditions(self) -> np.ndarray:
        """
        Defines some default initial conditions for the model
        """
        pass

    def __call__(self,
                 ts: np.ndarray,  # Array of time points, also determines step size of RK4 solver
                 light_est: np.ndarray,  # Array of light estimates, should be the same length as ts
                 state: np.ndarray,  # Initial state of the model
                 *args,
                 **kwargs):
        """ Wrapper to integrate_model, can just call the model object directly """
        return self.integrate_model(ts, light_est=light_est, state=state, *args, **kwargs)

    def initial_conditions_loop(self,
                                ts: np.ndarray, #Array of time points, also determines step size of RK4 solver
                                light_est: np.ndarray, #Array of light estimates, should be the same length as ts
                                num_loops: int = 30) -> np.ndarray: 
        """ 
            Estimate the starting values by looping the given light_estimate, commonly used for to estimate the initial conditions
            assumes the individual lives the same schedule repeatedly
        """
        ic = self.default_initial_conditions
        for _ in range(num_loops):
            sol = self.integrate_model(ts, light_est, ic).states
            ic = sol[..., -1]
        return ic


# %% ../nbs/00_models.ipynb 23
class Forger99Model(CircadianModel):
    """ Implementation of the Forger 1999 model """

    def __init__(self, params: dict= None):
        """
        Create a Forger VDP model 

        This will create a model with the default parameter values as given in Hannay et al 2019.

        This class can be used to simulate and plot the results of the given light schedule on the circadian phase
        and amplitude.
        """

        # Set the parameters to the published values by default
        self._default_params()
        
        if params:
            self.set_parameters(params)
        
        
    def _default_params(self):
        """
            Use the default parameters as defined in Hannay et al 2019
        """
        
        default_params = {'taux': 24.2,
                          'mu': 0.23,
                          'G': 33.75, 
                          'alpha_0': 0.05, 
                          'delta': 0.0075,
                          'p': 0.50, 
                          'I0': 9500.0, 
                          'kparam': 0.55}

        self.set_parameters(default_params)
        
    def set_parameters(self, param_dict: dict):
        """
            Update the model parameters using a passed in parameter dictionary. Any parameters not included
            in the dictionary will be set to the default values.

            updateParameters(param_dict)

            Returns null, changes the parameters stored in the class instance
        """
        

        params = [
            'taux',
            'mu'
            'G',
            'alpha_0',
            'delta',
            'p',
            'I0', 
            'kparam']

        for key, value in param_dict.items():
            setattr(self, key, value)
            
    def get_parameters_array(self):
        """
            Return a numpy array of the models current parameters
        """
        return np.array([self.taux, 
                         self.mu, 
                         self.G, 
                         self.alpha_0, 
                         self.delta, 
                         self.p, 
                         self.I0, 
                         self.kparam ])


    def get_parameters(self):
        """Get a dictionary of the current parameters being used by the model object.

        getParameters()

        returns a dict of parameters
        """

        current_params = {
            'taux': self.taux,
            'mu': self.mu,
            'G': self.G,
            'alpha_0': self.alpha_0,
            'delta': self.delta,
            'p': self.p,
            'I0': self.I0, 
            'kparam': self.kparam}

        return (current_params)

    def alpha0(self, 
               light: float # the light value in lux
               ):
        """A helper function for modeling the light input processing"""
        return (self.alpha_0 * pow((light / self.I0), self.p))

    def derv(self, 
             y: np.ndarray, # dynamical state (x, xc, n)
             light: float # light value in lux
             ) -> np.ndarray: # This defines the ode system for the forger 99 model

        x = y[...,0]
        xc = y[...,1]
        n = y[...,2]

        Bhat = self.G * (1.0 - n) * self.alpha0(light=light) * \
            (1 - 0.4 * x) * (1 - 0.4 * xc)

        dydt = np.zeros_like(y)
        dydt[...,0] = np.pi / 12.0 * (xc + Bhat)
        dydt[...,1] = np.pi / 12.0 * (self.mu * (xc - 4.0 / 3.0 * pow(xc, 3.0)) - x * (
            pow(24.0 / (0.99669 * self.taux), 2.0) + self.kparam * Bhat))
        dydt[...,2] = 60.0 * (self.alpha0(light=light) * (1.0 - n) - self.delta * n)

        return (dydt)
    
    def cbt(self, trajectory: DynamicalTrajectory) -> np.ndarray:
        cbt_mins = find_peaks(-1*trajectory.states[0,:])[0] # min of x is the CBTmin
        return trajectory.ts[cbt_mins]
    
    def dlmos(self, trajectory: DynamicalTrajectory) -> np.ndarray:
        return self.cbt(trajectory) - 7.0 # dlmo is defines by a relationship to cbt for this model

    def amplitude(state) -> float:
        return np.sqrt(state[0]**2+state[1]**2)

    def phase(state) -> float:
        x= state[0] 
        y = state[1]*-1.0
        return np.angle(x + complex(0,1)*y)
        
    @property
    def default_initial_conditions(self) -> np.ndarray:
        """
        x= –0.3 and xc= –1.13 are the default initial conditions for the model
        should be the value near the habitual bed time of the individual. 
        """
        return np.array([-0.3,-1.13,0.0])
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "Forger99Model"
        

# %% ../nbs/00_models.ipynb 38
class Hannay19TP(CircadianModel):
    """  The Hannay et al 2019 two population model, which models the ventral and dorsal SCN populations """

    def __init__(self, 
                 params: dict= None # Dict of parameters to set, the published values are used by default
                 ) -> None:
        # Set the parameters to the published values by default
        self._default_params()
        if params:
            self.set_parameters(params)
        
    def _default_params(self) -> None:
        """
            Stores the default parameters for the model as published in Hannet et al 2019
        """
        
        default_params = {'tauV': 24.25,
                          'tauD': 24.0,
                          'Kvv': 0.05, 
                          'Kdd': 0.04,
                          'Kvd': 0.05,
                          'Kdv': 0.01,
                          'gamma': 0.024,
                          'A1': 0.440068, 
                          'A2': 0.159136,
                          'BetaL': 0.06452, 
                          'BetaL2': -1.38935, 
                          'sigma': 0.0477375,
                          'G': 33.75, 
                          'alpha_0': 0.05, 
                          'delta': 0.0075,
                          'p': 1.5, 
                          'I0': 9325.0}

        self.set_parameters(default_params)
        
    def set_parameters(self, 
                       param_dict: dict # Dict of parameters to set
                       ) -> None:

        params = [
            'tauV',
            'tauD',
            'Kvv',
            'Kdd',
            "Kvd",
            "Kdv",
            'gamma',
            'A1',
            'A2',
            'BetaL',
            'BetaL2',
            'sigma',
            'G',
            'alpha_0',
            'delta',
            'p',
            'I0']

        for key, value in param_dict.items():
            setattr(self, key, value)
            
    def get_parameters_array(self):
        """
            Return a numpy array of the models current parameters
        """
        return np.array([self.tauV, self.tauD, self.Kvv, self.Kdd, self.Kvd, self.Kdv, self.gamma, self.BetaL, self.BetaL2, self.A1, self.A2, self.sigma, self.G, self.alpha_0, self.delta, self.p, self.I0 ])

    @property
    def get_parameters(self) -> dict:
    
        current_params = {
            'tauV': self.w0,
            'tauD': self.tauD,
            'Kvv': self.Kvv,
            'Kdd': self.Kdd,
            'Kdv': self.Kdv,
            'Kvd': self.Kdv,
            'gamma': self.gamma,
            'A1': self.A1,
            'A2': self.A2,
            'BetaL': self.BetaL,
            'BetaL2': self.BetaL2,
            'sigma': self.sigma,
            'G': self.G,
            'alpha_0': self.alpha_0,
            'delta': self.delta,
            'p': self.p,
            'I0': self.I0}

        return (current_params)

    def alpha0(self, 
               light: float # light intensity in lux
               ) -> float:
        """A helper function for modeling the light input processing"""
        return (self.alpha_0 * pow(light, self.p) /
                (pow(light, self.p) + self.I0))

    def derv(self, 
             y: np.ndarray, # state vector for the dynamical system (Rv, Rd, Psiv, Psid, n)
             light: float, # light intensity in lux
             ) -> np.ndarray: # derivative of the state vector

        Rv = y[...,0]
        Rd = y[...,1]
        Psiv = y[...,2]
        Psid = y[...,3]
        n = y[...,4]

        Bhat = self.G * (1.0 - n) * self.alpha0(light=light)

        LightAmp = self.A1 * 0.5 * Bhat * (1.0 - pow(Rv, 4.0)) * np.cos(Psiv + self.BetaL) + self.A2 * 0.5 * Bhat * Rv * (
            1.0 - pow(Rv, 8.0)) * np.cos(2.0 * Psiv + self.BetaL2)
        LightPhase = self.sigma * Bhat - self.A1 * Bhat * 0.5 * (pow(Rv, 3.0) + 1.0 / Rv) * np.sin(
            Psiv + self.BetaL) - self.A2 * Bhat * 0.5 * (1.0 + pow(Rv, 8.0)) * np.sin(2.0 * Psiv + self.BetaL2)

        dydt = np.zeros_like(y)
        dydt[...,0] = -self.gamma * Rv + self.Kvv / 2.0 * Rv * (1 - pow(Rv, 4.0)) + self.Kdv / 2.0 * Rd * (
            1 - pow(Rv, 4.0)) * np.cos(Psid - Psiv) + LightAmp
        dydt[...,1] = -self.gamma * Rd + self.Kdd / 2.0 * Rd * \
            (1 - pow(Rd, 4.0)) + self.Kvd / 2.0 * Rv * (1.0 - pow(Rd, 4.0)) * np.cos(Psid - Psiv)
        dydt[...,2] = 2.0 * np.pi / self.tauV + self.Kdv / 2.0 * Rd * \
            (pow(Rv, 3.0) + 1.0 / Rv) * np.sin(Psid - Psiv) + LightPhase
        dydt[...,3] = 2.0 * np.pi / self.tauD - self.Kvd / 2.0 * \
            Rv * (pow(Rd, 3.0) + 1.0 / Rd) * np.sin(Psid - Psiv)
        dydt[...,4] = 60.0 * (self.alpha0(light=light) * (1.0 - n) - self.delta * n)
        return dydt
    
    def dlmos(self, trajectory: DynamicalTrajectory) -> np.ndarray:
        return self.observer(trajectory, Hannay19TP.DLMOObs)
    
    def cbt(self, trajectory: DynamicalTrajectory) -> np.ndarray:
        return self.observer(trajectory, Hannay19TP.CBTObs)

    def observer(self, trajectory: np.ndarray, observer_func: callable) -> np.array:
        zero_crossings = np.where(np.diff(np.sign(observer_func(0.0, trajectory.states))))[0]
        return trajectory.ts[zero_crossings]

    def DLMOObs(t, state) -> float:
        return np.sin(0.5*(Hannay19TP.phase(state)-5*np.pi/12.0))
    
    def CBTObs(t, state) -> float:
        return np.sin(0.5*(Hannay19TP.phase(state)-np.pi))

    def amplitude(state):
        # Make this joint amplitude at some point 
        return(state[0])

    def phase(state):
        return(state[2])
    
    @staticmethod
    def phase_difference(state):
        return state[2] - state[3]
    
    @property
    def default_initial_conditions(self) -> np.ndarray:
        return np.array([1.0,1.0,0.0,0.10,0.0])
    




# %% ../nbs/00_models.ipynb 41
class Hannay19(CircadianModel):
    """
        A simple python program to integrate the human circadian rhythms model 
        (Hannay et al 2019) for a given light schedule
    """

    def __init__(self, params: dict = None):
        """
            Create a single population model by passing in a Light Function as a function of time.

            This will create a model with the default parameter values as given in Hannay et al 2019.

            This class can be used to simulate and plot the results of the given light schedule on the circadian phase
            and amplitude.
        """
        self._default_params()
        if params:
            self.set_parameters(params)

    def _default_params(self):
        """
            Use the default parameters as defined in Hannay et al 2019
        """
        default_params = {'tau': 23.84, 'K': 0.06358, 'gamma': 0.024,
                          'Beta1': -0.09318, 'A1': 0.3855, 'A2': 0.1977,
                          'BetaL1': -0.0026, 'BetaL2': -0.957756, 'sigma': 0.0400692,
                          'G': 33.75, 'alpha_0': 0.05, 'delta': 0.0075,
                          'p': 1.5, 'I0': 9325.0}

        self.set_parameters(default_params)

    def set_parameters(self, param_dict: dict):
        """
            Update the model parameters using a passed in parameter dictionary. Any parameters not included
            in the dictionary will be set to the default values.

            updateParameters(param_dict)

            Returns null, changes the parameters stored in the class instance
        """

        params = ['tau', 'K', 'gamma', 'Beta1', 'A1', 'A2', 'BetaL1',
                  'BetaL2', 'sigma', 'G', 'alpha_0', 'delta', 'p', 'I0']

        for key, value in param_dict.items():
            setattr(self, key, value)

    def get_parameters(self):
        """
            Get a dictionary of the current parameters being used by the model object.

                get_parameters()

            returns a dict of parameters
        """

        current_params = {'tau': self.tau, 'K': self.K, 'gamma': self.gamma,
                          'Beta1': self.Beta1, 'A1': self.A1, 'A2': self.A2,
                          'BetaL1': self.BetaL1,
                          'BetaL2': self.BetaL2, 'sigma': self.sigma,
                          'G': self.G, 'alpha_0': self.alpha_0,
                          'delta': self.delta, 'p': self.p, 'I0': self.I0}

        return(current_params)

    def get_parameters_array(self):
        """
            Return a numpy array of the models current parameters
        """
        return np.array([self.tau, self.K, self.gamma, self.Beta1, self.A1, self.A2, self.BetaL1, self.BetaL2, self.sigma, self.G, self.alpha_0, self.delta, self.I0, self.p])
    
    def alpha0(self, light: float):
        """A helper function for modeling the light input processing"""
        return (self.alpha_0 * pow(light, self.p) /
                (pow(light, self.p) + self.I0))
    
    def derv(self, 
             y: np.ndarray, # circadian state where the last dimension is the state variable
             light: float, # light level in lux 
             ):
        R = y[...,0]
        Psi = y[..., 1]
        n = y[...,2]

        Bhat = self.G * (1.0 - n) * self.alpha0(light=light)
        LightAmp = self.A1 * 0.5 * Bhat * (1.0 - pow(R, 4.0)) * np.cos(Psi + self.BetaL1) + self.A2 * 0.5 * Bhat * R * (
            1.0 - pow(R, 8.0)) * np.cos(2.0 * Psi + self.BetaL2)
        LightPhase = self.sigma * Bhat - self.A1 * Bhat * 0.5 * (pow(R, 3.0) + 1.0 / R) * np.sin(
            Psi + self.BetaL1) - self.A2 * Bhat * 0.5 * (1.0 + pow(R, 8.0)) * np.sin(2.0 * Psi + self.BetaL2)

        dydt = np.zeros_like(y)
        dydt[...,0] = -1.0 * self.gamma * R + self.K * \
            np.cos(self.Beta1) / 2.0 * R * (1.0 - pow(R, 4.0)) + LightAmp
        dydt[...,1] = 2*np.pi/self.tau + self.K / 2.0 * \
            np.sin(self.Beta1) * (1 + pow(R, 4.0)) + LightPhase
        dydt[...,2] = 60.0 * (self.alpha0(light=light) * (1.0 - n) - self.delta * n)

        return (dydt)
        

    def integrate_observer(self, ts: np.ndarray, light_est: np.ndarray, u0: np.ndarray = None, observer=None):
        """
            Integrate the spmodel forward in time using the given light estimate vector
        """
        if observer is None:
            observer = Hannay19.DLMOObs
        sol = self.integrate_model(ts, light_est, u0)
        zero_crossings = np.where(np.diff(np.sign(observer(0.0, sol))))[0]
        return ts[zero_crossings]
    
    def dlmos(self, trajectory: DynamicalTrajectory) -> np.ndarray:
        return self.observer(trajectory, Hannay19.DLMOObs)
    
    def cbt(self, trajectory: DynamicalTrajectory) -> np.ndarray:
        return self.observer(trajectory, Hannay19.CBTObs)

    def observer(self, trajectory: np.ndarray, observer_func: callable) -> np.array:
        zero_crossings = np.where(np.diff(np.sign(observer_func(0.0, trajectory.states)), axis=-1))[0]
        return trajectory.ts[zero_crossings]

    def DLMOObs(t, state) -> float:
        return np.sin(0.5*(Hannay19.phase(state)-5*np.pi/12.0))
    
    def CBTObs(t, state) -> float:
        return np.sin(0.5*(Hannay19.phase(state)-np.pi))

    def amplitude(state) -> float:
        return(state[0])

    def phase(state) -> float:
        return(state[1])
    
    @property
    def default_initial_conditions(self) -> np.ndarray:
        """
        Gives some default initial conditions for the model
        """
        return np.array([0.70,0.0,0.0])




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_sleep.ipynb.

# %% auto 0
__all__ = ['TwoProcessModel', 'sleep_midpoint', 'cluster_sleep_periods_scipy']

# %% ../nbs/05_sleep.ipynb 2
from nbdev import show_doc

# %% ../nbs/05_sleep.ipynb 3
import numpy as np
import pandas as pd
import scipy as sp
from numba import jit
from scipy import interpolate
import torch 
import json
import pytz
import datetime
import copy
import scipy as sp
from scipy.integrate import solve_ivp
from math import *
import pylab as plt

from scipy.optimize import minimize
from .utils import *
from .models import * 
from .lights import *
from .plots import *

# %% ../nbs/05_sleep.ipynb 5
class TwoProcessModel:

    def __init__(self, steps_wake_threshold: float = 10.0):
        self.steps_wake_threshold = steps_wake_threshold
        self.awake = True

    @staticmethod
    def check_wake_status(awake: bool, # current wake status
                          h: float,  # homeostat value
                          phase: float #circadian phase
                          ) -> bool:

        c = np.cos(phase)
        H_minus = 0.17
        H_plus = 0.6
        homeostat_a = 0.10

        upper = (H_plus + homeostat_a * c)
        lower = (H_minus + homeostat_a * c)
        above_threshold = h > upper
        below_threshold = h <= lower

        if above_threshold:
            return False
        elif below_threshold:
            return True
        else:
            return awake

    def dhomeostat(self, 
                   homeostat: float, # homeostat value
                   steps: float, # steps value
                   phase: float, #circadian phase
                   ) -> float:

        h = homeostat[0]
        tau_s = 4.2  # hours
        tau_w = 18.2  # hours
        mu_s = 1
        
        step_awake = (steps > self.steps_wake_threshold) or self.awake
        dH = (mu_s - h) / tau_w if step_awake else -h / tau_s
        self.awake = TwoProcessModel.check_wake_status(
            self.awake, 
            h, 
            phase
            )
        return np.array([dH]) 
    
    # implement a RK4 solver
    def step_rk4(self,
                 state: np.ndarray,
                 steps: float,
                 phase: float,
                 dt=0.10):

        k1 = self.dhomeostat(state, steps, phase) 
        k2 = self.dhomeostat(state + k1 * dt / 2.0, steps, phase)
        k3 = self.dhomeostat(state + k2 * dt / 2.0, steps, phase)
        k4 = self.dhomeostat(state + k3 * dt, steps, phase)
        state = state + (dt / 6.0) * (k1 + 2.0*k2 + 2.0*k3 + k4)
        return state

    def __call__(self, 
                 ts: np.ndarray,
                 phase: np.ndarray,
                 steps: np.ndarray,
                 initial_value: np.array = np.array([0.50]) #  initial value for the homeostat
                 ) -> DynamicalTrajectory:
        sol = np.zeros((len(initial_value), len(ts)))
        current_state = initial_value
        sol[:,0] = current_state
        for idx in range(1,len(ts)):
            current_state = self.step_rk4(current_state, steps[idx], phase[idx])
            sol[:,idx] = current_state
        return(DynamicalTrajectory(ts, np.array(sol)))

# %% ../nbs/05_sleep.ipynb 7
def sleep_midpoint(timetotal: np.ndarray, 
                   Wake: np.ndarray, 
                   durations=True):
    """
        Given a wearable data frame with a Wake column which takes 
        the values 0, 1, missing this routine will create a sleep phase 
        column which is based on constant phase accumulation between sleep 
        midpoints. 

        The sleep midpoints are found using the criteria than they the median 
        time where 

    """

    sleep_start = []
    sleep_end = []
    awake = Wake[0] > 0.50

    if not awake:
        sleep_start.append(timetotal[1])

    for k in range(1, len(Wake)):
        if (Wake[k] > 0.50 and not awake):
            awake = True
            sleep_end.append(timetotal[k])

        if (Wake[k] <= 0.50 and awake):
            awake = False
            sleep_start.append(timetotal[k])

    if Wake[-1] <= 0.50:
        sleep_end.append(timetotal[-1])

    assert len(sleep_start) == len(sleep_end)
    sleep_midpoints = []
    sleep_durations = []
    for (s1, s2) in zip(sleep_start, sleep_end):
        sleep_midpoints += [(s2-s1)/2+s1]
        sleep_durations += [s2-s1]

    if durations:
        return np.array(sleep_midpoints), np.array(sleep_durations)
    else:
        return np.array(sleep_midpoints)



# %% ../nbs/05_sleep.ipynb 8
def cluster_sleep_periods_scipy(wake_data: np.ndarray, 
                                epsilon: float,
                                makeplot: bool = False,
                                max_sleep_clusters=None, 
                                min_sleep_clusters=None):
    """
        Given a binary vector wake_data which gives a prediction for the sleep/wake  
        status and a regularization penalty ε this function will create smoothed 
        sleep-wake periods. This helps as preprocessing to remove erroneous short sleep 
        periods (and wake) which may mess up calculations like the sleep midpoint for 
        the day

        cluster_sleep_periods(wake_data : np.ndarray, epsilon: float, makeplot: bool=False):
    """

    np.nan_to_num(wake_data, 0.50)

    def objective(w):
        return sum(w * (1 - wake_data)) + sum((1 - w)*wake_data) + epsilon*sum((w[1:]-w[0:-1])**2)

    max_clusters = max_sleep_clusters or len(wake_data)

    def constraint1(x):
        return max_clusters-sum((x[1:]-x[0:-1])**2)  # geq 0

    min_clusters = min_sleep_clusters or 0

    print(
        f"The max clusters are {max_clusters} and the min clusters are {min_clusters}")

    def constraint2(x):
        return sum((x[1:]-x[0:-1])**2)-min_clusters  # geq 0

    bnds = (0.0, 1.0)
    all_bnds = [bnds for b in range(len(wake_data))]

    constraint1d = {'type': 'ineq', 'fun': constraint1}
    constraint2d = {'type': 'ineq', 'fun': constraint2}
    all_cons = [constraint1d, constraint2d]

    x0 = wake_data
    sol = minimize(objective, x0, method='SLSQP', bounds=all_bnds)

    if makeplot:
        pl = plt.scatter(range(len(wake_data)), wake_data + 0.1 *
                         np.random.randn(len(wake_data)), label="", color="blue")
        plt.plot(range(len(wake_data)), convert_binary(
            sol.x), lw=2.0, label="", color="red")
        plt.show()

    print(
        f"The max clusters are {max_clusters} takes value {constraint1(sol.x)}>=0.0")
    print(
        f"The min clusters are {min_clusters} and takes the value {constraint2(sol.x)}>=0.0")
    return(convert_binary(sol.x))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_plots.ipynb.

# %% auto 0
__all__ = ['Actogram', 'plot_mae', 'plot_torus', 'Stroboscopic']

# %% ../nbs/03_plots.ipynb 3
from scipy.ndimage import gaussian_filter1d
from copy import deepcopy
import pylab as plt
import numpy as np
from pathlib import Path
import sys
from .utils import cut_phases_12
from .models import Hannay19
from .lights import *


# %% ../nbs/03_plots.ipynb 5
class Actogram:
    """ 
        Create an Actogram visualisation of the data
    """

    def __init__(self,
                 time_total: np.ndarray, # time in hours
                 light_vals: np.ndarray, # light values or proxy should be the same length as time_total
                 second_zeit: np.ndarray = None, # optional second zeitgeber to show on the right side
                 ax=None, # Axis to plot on, if None a new figure is created
                 threshold=10.0, # threshold for light on/off 
                 threshold2=None, # threshold for light on/off for second zeitgeber
                 opacity: float = 1.0, # opacity of the light on/off rectangles
                 color: str = 'black', # color of the light on/off rectangles, for the below threshold values
                 smooth=True, # whether to apply a gaussian filter to the light values to smooth it, useful for wearable data
                 sigma=[2.0, 2.0], # sigma for the gaussian filter
                 *args, # additional plot arguments
                 **kwargs # additional plot keyword arguments
                 ):

        self.time_total = time_total
        self.light_vals = light_vals
        self.num_days = np.ceil((time_total[-1] - time_total[0])/24.0)

        self.second_zeit = second_zeit if second_zeit is not None else light_vals

        if smooth:
            self.light_vals = gaussian_filter1d(
                self.light_vals, sigma=sigma[0])
            self.second_zeit = gaussian_filter1d(
                self.second_zeit, sigma=sigma[1])

        threshold2 = threshold if threshold2 is None else threshold2
        self.ax = ax if ax is not None else plt.gca()

        # Set graphical parameters
        label_scale = int(np.floor(self.num_days/30))
        if label_scale < 1:
            label_scale = 1

        self.opacity = opacity
        self.darkColor = color
        start_day = np.floor(self.time_total[0]/24.0)
        self.ax.set_ylim(start_day, self.num_days+start_day)
        self.ax.set_xlim(0, 48)
        self.ax.set_yticks(
            np.arange(int(start_day), self.num_days+1+start_day, label_scale))
        ylabels_list = list(range(int(start_day), int(
            self.num_days+start_day)+1, label_scale))
        # ylabels_list.reverse()

        self.ax.set_yticklabels(ylabels_list)
        self.ax.set_xticks(np.arange(0, 48+3, 3))
        xlabels_list = list(range(0, 27, 3))+list(range(3, 27, 3))
        self.ax.set_xticklabels(xlabels_list)
        self.ax.set_xticks(np.arange(0, 48, 1), minor=True)

        #self.ax.yaxis.grid(False, linewidth=1.0, color='k')
        # self.ax.xaxis.grid(False)

        self.ax.plot(24.0*np.ones(100), 
                     np.linspace(0, self.num_days,100), 
                     ls='--', 
                     lw=2.0, 
                     color='black', 
                     zorder=9)
        self.ax.set_xlabel("ZT")
        self.ax.set_ylabel("Days")

        self.addLightSchedule(self.light_vals, threshold,
                              plt_option='left', color=self.darkColor)
        self.addLightSchedule(self.second_zeit, threshold2,
                              plt_option='right', color=self.darkColor)
        self.ax.invert_yaxis()

    def getRectangles(self, timeon, timeoff, colorIn='white'):
        bottom_x = np.fmod(timeon, 24.0)
        bottom_y = int(timeon/24.0)  # -1
        alpha = self.opacity if colorIn != 'white' else 0.0
        r1 = plt.Rectangle((bottom_x, bottom_y), timeoff -
                           timeon, 1, fc=colorIn, zorder=-1, alpha=alpha)
        r2 = plt.Rectangle((bottom_x+24.0, bottom_y),
                           timeoff-timeon, 1, fc=colorIn, zorder=1, alpha=alpha)
        return((r1, r2))

    def addRect(self, timeon, timeoff, colorIn='white', plt_option='both'):
        """Used to add a rectangle to the axes"""
        r = self.getRectangles(timeon, timeoff, colorIn)
        if plt_option == 'left':
            self.ax.add_patch(r[0])
            return
        if plt_option == 'right':
            self.ax.add_patch(r[1])
            return
        self.ax.add_patch(r[0])
        self.ax.add_patch(r[1])

    def addLightSchedule(self, zeit: np.ndarray,
                         threshold: float,
                         plt_option: str = 'both',
                         color='black'):
        lightdata = zeit
        timedata = self.time_total
        lightsOn = False
        if (lightdata[0] > threshold):
            lightsOn = True
            lightStart = timedata[0]
        else:
            darkOn = True
            darkStart = timedata[0]

        dayCounter = int(timedata[0]/24.0)  # count the days in the data set
        for i in range(1, len(lightdata)):
            currentDay = int(timedata[i]/24.0)
            if (currentDay != dayCounter):
                dayCounter = currentDay
                if (lightsOn == True):
                    self.addRect(
                        lightStart, timedata[i], plt_option=plt_option)
                    if (i+1 < len(timedata)):
                        # reset the light counter to start over the next day
                        lightStart = timedata[i+1]
                else:
                    self.addRect(
                        darkStart, timedata[i], colorIn=color, plt_option=plt_option)
                    if (i+1 < len(timedata)):
                        darkStart = timedata[i+1]

            if (lightdata[i] < threshold and lightsOn == True):
                self.addRect(lightStart, timedata[i-1], plt_option=plt_option)
                lightsOn = False
                darkOn = True
                darkStart = timedata[i]
            if (lightsOn == False and lightdata[i] >= threshold):
                lightsOn = True
                lightStart = timedata[i]
                darkOn = False
                self.addRect(
                    darkStart, timedata[i-1], colorIn=color, plt_option=plt_option)

    def plot_phasemarker(self, phase_marker_times: np.ndarray,
                         error: np.ndarray = None,
                         alpha=1.0,
                         alpha_error=0.30,
                         scatter=False,
                         *args, **kwargs):
        """
        This method takes in a list of times which are assumed to occur at the same 
        circadian phase (e.g. DLMO, CBTmin). These are plotted as points 
        on the actogram.  
            plot_phasemarker(self, phase_marker_times: np.ndarray, *args, **kwargs)
        """

        xvals = deepcopy(phase_marker_times)
        yvals = deepcopy(phase_marker_times)

        xvals = np.fmod(xvals, 24.0)
        yvals = np.floor(yvals / 24.0) + 0.5

        if scatter:
            self.ax.scatter(xvals, yvals, *args, **kwargs)
            self.ax.scatter(xvals+24.0, yvals, *args, **kwargs)

        idx_split = (np.absolute(np.diff(xvals)) > 6.0).nonzero()[0]+1
        xvals_split = np.split(xvals, idx_split)
        yvals_split = np.split(yvals, idx_split)
        if error is not None:
            error_split = np.split(error, idx_split)

        for (idx, xx) in enumerate(xvals_split):
            self.ax.plot(xx, yvals_split[idx], alpha=alpha, *args, **kwargs)
            self.ax.plot(
                xx+24.0, yvals_split[idx], alpha=alpha, *args, **kwargs)
            if error is not None:
                self.ax.fill_betweenx(
                    yvals_split[idx], xx-error_split[idx], xx+error_split[idx], alpha=alpha_error, *args, **kwargs)
                self.ax.fill_betweenx(
                    yvals_split[idx], xx-error_split[idx]+24.0, xx+error_split[idx]+24.0, alpha=alpha_error, *args, **kwargs)

    def plot_phasetimes(self, times: np.ndarray, phases: np.ndarray, error: np.ndarray = None,
                        alpha_error=0.30, alpha=1.0, *args, **kwargs):
        """
            This method takes observations of the phase and times (same length arrays)
            and adds them to the actogram.    

            plot_phasetimes(self, times: np.ndarray, phases: np.ndarray, *args, **kwargs)
        """
        xvals = deepcopy(phases)
        xvals = np.arctan2(np.sin(xvals), np.cos(xvals))
        for i in range(len(xvals)):
            if xvals[i] < 0.0:
                xvals[i] += 2*np.pi

        xvals = np.fmod(xvals, 2*np.pi)
        xvals *= 12.0/np.pi

        xvals = np.fmod(xvals, 24.0)
        yvals = deepcopy(times)
        yvals = np.floor(yvals / 24.0) + 0.5

        idx_split = (np.absolute(np.diff(xvals)) > 6.0).nonzero()[0]+1
        xvals_split = np.split(xvals, idx_split)
        yvals_split = np.split(yvals, idx_split)
        if error is not None:
            error_split = np.split(error, idx_split)

        for (idx, xx) in enumerate(xvals_split):
            self.ax.plot(xx, yvals_split[idx], alpha=alpha, *args, **kwargs)
            self.ax.plot(
                xx+24.0, yvals_split[idx], alpha=alpha, *args, **kwargs)
            if error is not None:
                self.ax.fill_betweenx(
                    yvals_split[idx], xx-error_split[idx], xx+error_split[idx], alpha=alpha_error, *args, **kwargs)
                self.ax.fill_betweenx(
                    yvals_split[idx], xx-error_split[idx]+24.0, xx+error_split[idx]+24.0, alpha=alpha_error, *args, **kwargs)



# %% ../nbs/03_plots.ipynb 7
def plot_mae(dlmo_actual: np.ndarray,  # expected to be in hours
             dlmo_pred: np.ndarray,  # predicted to be in hours
             norm_to: float = None, 
             ax=None,  
             *args, 
             **kwargs):

    dlmo_actual = np.fmod(dlmo_actual, 24.0)
    dlmo_pred = np.fmod(dlmo_pred, 24.0)

    # Make the plot range from from -12 to 12
    dlmo_pred = np.array([cut_phases_12(d) for d in list(dlmo_pred)])
    dlmo_actual = np.array([cut_phases_12(d) for d in list(dlmo_actual)])

    if norm_to is not None:
        dlmo_actual = dlmo_actual - np.mean(dlmo_actual)+norm_to

    if ax is None:
        plt.figure()
        ax = plt.gca()

    errors = dlmo_pred-dlmo_actual
    print(f"The MAE is: {np.mean(abs(errors))}")
    print(f"Within one hour {np.sum(abs(errors)<=1.0)}/{len(dlmo_pred)}")
    print(errors)

    ax.scatter(dlmo_actual, dlmo_pred, *args, **kwargs)
    ax.plot(np.arange(-12, 12, 1), np.arange(-12, 12, 1),
            ls='--', color='gray', lw=2.0)
    ax.plot(np.arange(-12, 12, 1), np.arange(-12, 12, 1)+1,
            ls='--', color='blue', lw=1.0)
    ax.plot(np.arange(-12, 12, 1), np.arange(-12, 12, 1)-1,
            ls='--', color='blue', lw=1.0)

    ax.set_ylabel("Model Prediction (hrs)")
    ax.set_xlabel("Experimental DLMO (hrs)")



# %% ../nbs/03_plots.ipynb 9
def plot_torus(phase1: np.ndarray, # array of phases  
               phase2: np.ndarray, #array of phases, assumed to be the same length as phase1
               scaled_by: float = None, # should the phases be wrapped, this just applies an fmod to the phases
               ax=None, # axis to plot on, if None, a new figure is created
               *args, # passed to the scatter plot
               **kwargs, # passed to the scatter plot
               ) -> plt.Axes:

    if ax is None:
        plt.figure()
        ax = plt.gca()

    if scaled_by is not None:
        phase1 = np.fmod(phase1, scaled_by) * np.pi/12.0
        phase2 = np.fmod(phase2,  scaled_by) * np.pi/24.0

    phase1 = np.arctan2(np.sin(phase1), np.cos(phase1))
    phase2 = np.arctan2(np.sin(phase2), np.cos(phase2))

    ax.scatter(phase1, phase2, *args, **kwargs)
    
    return ax


# %% ../nbs/03_plots.ipynb 13
class Stroboscopic:
    """
    This class can be used to make a stroboscopic plot of the entrainment of an oscillator to a sudden shift in schedule
    """

    def __init__(self, 
                 ax: plt.Axes, # The axes to plot on
                 ts: np.ndarray, # The time series of the model
                amplitude: np.ndarray, # The amplitude of the model, assumed to be the same length as ts
                phase: np.ndarray, # The phase of the model, assumed to be the same length as ts
                period: float = 24.0, # The time period between stroboscopic arrows
                *args: tuple, #passed to the quiver plot
                **kwargs: dict #passed to the quiver plot
        ):
        
        self.ts = ts
        self.R = amplitude
        self.phase = phase
        self.period = period
        self.ax = ax

        self._make_strobo_plot(*args, **kwargs)

    def _make_strobo_plot(self, *args, **kwargs):
        """Add the paths to a quiver plot"""
        start_amp = self.R[0]

        Xvals = np.array(self.R / start_amp * np.cos(self.phase))
        Yvals = np.array(self.R / start_amp * np.sin(self.phase))

        circle_angles = np.linspace(0, 2 * np.pi, 1000)
        circle_x = list([np.cos(x) for x in circle_angles])
        circle_y = list([np.sin(x) for x in circle_angles])

        self.ax.plot(circle_x, circle_y, lw=2.0, color='k')
        # Sample down to every 24 hours
        Xvals = np.interp(np.arange(self.ts[0], self.ts[-1], self.period), self.ts, Xvals)
        Yvals = np.interp(np.arange(self.ts[0], self.ts[-1], self.period), self.ts, Yvals)
        upper_bound = len(Xvals)
        for i in range(1, upper_bound):
            self.ax.quiver(Xvals[i - 1], Yvals[i - 1], Xvals[i] - Xvals[i - 1], Yvals[i] - Yvals[i - 1],
                           scale_units='xy', angles='xy', scale=1, *args, **kwargs)
        self.ax.set_xlim([-1.1, 1.1])
        self.ax.set_ylim([-1.1, 1.1])
        self.ax.scatter([0.0], [0.0], color='k')
        self.ax.set_axis_off()

    def add_strobo_plot(self, 
                        amplitude: np.ndarray,
                        phase: np.ndarray,
                        col='darkgreen'):
        """Add a strobo plot to the axes for comparison"""
        start_amp = amplitude[0]

        Xvals = np.array(amplitude / start_amp * np.cos(phase))
        Yvals = np.array(amplitude / start_amp * np.sin(phase))

        # Sample down to every 24 hours
        Xvals = Xvals[::240]
        Yvals = Yvals[::240]

        upper_bound = min(10, len(Xvals))
        for i in range(1, upper_bound + 10):
            self.ax.quiver(Xvals[i - 1], Yvals[i - 1], Xvals[i] - Xvals[i - 1], Yvals[i] - Yvals[i - 1],
                           scale_units='xy', angles='xy', scale=1, color=col)




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/circadian',
                'doc_host': 'https://khannay.github.io',
                'git_url': 'https://github.com/khannay/circadian',
                'lib_path': 'circadian'},
  'syms': { 'circadian.cli': { 'circadian.cli.main_acto': ('cli.html#main_acto', 'circadian/cli.py'),
                               'circadian.cli.main_esri': ('cli.html#main_esri', 'circadian/cli.py')},
            'circadian.lights': { 'circadian.lights.DosageResponseCurve': ('lights.html#dosageresponsecurve', 'circadian/lights.py'),
                                  'circadian.lights.DosageResponseCurve.__init__': ( 'lights.html#dosageresponsecurve.__init__',
                                                                                     'circadian/lights.py'),
                                  'circadian.lights.DosageResponseCurve.light_dosage': ( 'lights.html#dosageresponsecurve.light_dosage',
                                                                                         'circadian/lights.py'),
                                  'circadian.lights.DosageResponseCurve.light_dosage_day1': ( 'lights.html#dosageresponsecurve.light_dosage_day1',
                                                                                              'circadian/lights.py'),
                                  'circadian.lights.IntensityResponseCurveLight': ( 'lights.html#intensityresponsecurvelight',
                                                                                    'circadian/lights.py'),
                                  'circadian.lights.IntensityResponseCurveLight.__init__': ( 'lights.html#intensityresponsecurvelight.__init__',
                                                                                             'circadian/lights.py'),
                                  'circadian.lights.IntensityResponseCurveLight.light_intensity': ( 'lights.html#intensityresponsecurvelight.light_intensity',
                                                                                                    'circadian/lights.py'),
                                  'circadian.lights.Light': ('lights.html#light', 'circadian/lights.py'),
                                  'circadian.lights.Light.RegularLight': ('lights.html#light.regularlight', 'circadian/lights.py'),
                                  'circadian.lights.Light.ShiftWorkLight': ('lights.html#light.shiftworklight', 'circadian/lights.py'),
                                  'circadian.lights.Light.SlamShift': ('lights.html#light.slamshift', 'circadian/lights.py'),
                                  'circadian.lights.Light.__add__': ('lights.html#light.__add__', 'circadian/lights.py'),
                                  'circadian.lights.Light.__call__': ('lights.html#light.__call__', 'circadian/lights.py'),
                                  'circadian.lights.Light.__init__': ('lights.html#light.__init__', 'circadian/lights.py'),
                                  'circadian.lights.Light.end_time': ('lights.html#light.end_time', 'circadian/lights.py'),
                                  'circadian.lights.Light.plot': ('lights.html#light.plot', 'circadian/lights.py'),
                                  'circadian.lights.PRCFinder': ('lights.html#prcfinder', 'circadian/lights.py'),
                                  'circadian.lights.PRCFinder.exp_type0': ('lights.html#prcfinder.exp_type0', 'circadian/lights.py'),
                                  'circadian.lights.PRCFinder.prc_type0_point': ( 'lights.html#prcfinder.prc_type0_point',
                                                                                  'circadian/lights.py'),
                                  'circadian.lights.PRCFinder.type0x': ('lights.html#prcfinder.type0x', 'circadian/lights.py'),
                                  'circadian.lights.PhaseResponseCurveLight': ( 'lights.html#phaseresponsecurvelight',
                                                                                'circadian/lights.py'),
                                  'circadian.lights.PhaseResponseCurveLight.__init__': ( 'lights.html#phaseresponsecurvelight.__init__',
                                                                                         'circadian/lights.py'),
                                  'circadian.lights.PhaseResponseCurveLight.light_amplitude_reseting': ( 'lights.html#phaseresponsecurvelight.light_amplitude_reseting',
                                                                                                         'circadian/lights.py'),
                                  'circadian.lights.PhaseResponseCurveLight.light_czeiler_type0': ( 'lights.html#phaseresponsecurvelight.light_czeiler_type0',
                                                                                                    'circadian/lights.py'),
                                  'circadian.lights.PhaseResponseCurveLight.light_hilaire': ( 'lights.html#phaseresponsecurvelight.light_hilaire',
                                                                                              'circadian/lights.py'),
                                  'circadian.lights.PhaseResponseCurveLight.light_khalsa': ( 'lights.html#phaseresponsecurvelight.light_khalsa',
                                                                                             'circadian/lights.py'),
                                  'circadian.lights.RimmerLightPulseLight': ('lights.html#rimmerlightpulselight', 'circadian/lights.py'),
                                  'circadian.lights.RimmerLightPulseLight.__init__': ( 'lights.html#rimmerlightpulselight.__init__',
                                                                                       'circadian/lights.py'),
                                  'circadian.lights.RimmerLightPulseLight.make_pulse_rimmer': ( 'lights.html#rimmerlightpulselight.make_pulse_rimmer',
                                                                                                'circadian/lights.py'),
                                  'circadian.lights.RimmerLightPulseLight.pulse_rimmer_start': ( 'lights.html#rimmerlightpulselight.pulse_rimmer_start',
                                                                                                 'circadian/lights.py'),
                                  'circadian.lights.get_pulse': ('lights.html#get_pulse', 'circadian/lights.py'),
                                  'circadian.lights.heaviside': ('lights.html#heaviside', 'circadian/lights.py'),
                                  'circadian.lights.make_pulse': ('lights.html#make_pulse', 'circadian/lights.py')},
            'circadian.metrics': { 'circadian.metrics.esri': ('metrics.html#esri', 'circadian/metrics.py'),
                                   'circadian.metrics.esri_trajectory': ('metrics.html#esri_trajectory', 'circadian/metrics.py')},
            'circadian.models': { 'circadian.models.CircadianModel': ('models.html#circadianmodel', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.__call__': ( 'models.html#circadianmodel.__call__',
                                                                                'circadian/models.py'),
                                  'circadian.models.CircadianModel.__init__': ( 'models.html#circadianmodel.__init__',
                                                                                'circadian/models.py'),
                                  'circadian.models.CircadianModel._default_params': ( 'models.html#circadianmodel._default_params',
                                                                                       'circadian/models.py'),
                                  'circadian.models.CircadianModel.amplitude': ( 'models.html#circadianmodel.amplitude',
                                                                                 'circadian/models.py'),
                                  'circadian.models.CircadianModel.cbt': ('models.html#circadianmodel.cbt', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.default_initial_conditions': ( 'models.html#circadianmodel.default_initial_conditions',
                                                                                                  'circadian/models.py'),
                                  'circadian.models.CircadianModel.dlmos': ('models.html#circadianmodel.dlmos', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.get_parameters': ( 'models.html#circadianmodel.get_parameters',
                                                                                      'circadian/models.py'),
                                  'circadian.models.CircadianModel.initial_conditions_loop': ( 'models.html#circadianmodel.initial_conditions_loop',
                                                                                               'circadian/models.py'),
                                  'circadian.models.CircadianModel.integrate_model': ( 'models.html#circadianmodel.integrate_model',
                                                                                       'circadian/models.py'),
                                  'circadian.models.CircadianModel.observer': ( 'models.html#circadianmodel.observer',
                                                                                'circadian/models.py'),
                                  'circadian.models.CircadianModel.phase': ('models.html#circadianmodel.phase', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.step_rk4': ( 'models.html#circadianmodel.step_rk4',
                                                                                'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory': ('models.html#dynamicaltrajectory', 'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__call__': ( 'models.html#dynamicaltrajectory.__call__',
                                                                                     'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__getitem__': ( 'models.html#dynamicaltrajectory.__getitem__',
                                                                                        'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__init__': ( 'models.html#dynamicaltrajectory.__init__',
                                                                                     'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__len__': ( 'models.html#dynamicaltrajectory.__len__',
                                                                                    'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.batch_size': ( 'models.html#dynamicaltrajectory.batch_size',
                                                                                       'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.get_batch': ( 'models.html#dynamicaltrajectory.get_batch',
                                                                                      'circadian/models.py'),
                                  'circadian.models.Forger99Model': ('models.html#forger99model', 'circadian/models.py'),
                                  'circadian.models.Forger99Model.__init__': ('models.html#forger99model.__init__', 'circadian/models.py'),
                                  'circadian.models.Forger99Model.__repr__': ('models.html#forger99model.__repr__', 'circadian/models.py'),
                                  'circadian.models.Forger99Model.__str__': ('models.html#forger99model.__str__', 'circadian/models.py'),
                                  'circadian.models.Forger99Model._default_params': ( 'models.html#forger99model._default_params',
                                                                                      'circadian/models.py'),
                                  'circadian.models.Forger99Model.alpha0': ('models.html#forger99model.alpha0', 'circadian/models.py'),
                                  'circadian.models.Forger99Model.amplitude': ( 'models.html#forger99model.amplitude',
                                                                                'circadian/models.py'),
                                  'circadian.models.Forger99Model.cbt': ('models.html#forger99model.cbt', 'circadian/models.py'),
                                  'circadian.models.Forger99Model.default_initial_conditions': ( 'models.html#forger99model.default_initial_conditions',
                                                                                                 'circadian/models.py'),
                                  'circadian.models.Forger99Model.derv': ('models.html#forger99model.derv', 'circadian/models.py'),
                                  'circadian.models.Forger99Model.dlmos': ('models.html#forger99model.dlmos', 'circadian/models.py'),
                                  'circadian.models.Forger99Model.get_parameters': ( 'models.html#forger99model.get_parameters',
                                                                                     'circadian/models.py'),
                                  'circadian.models.Forger99Model.get_parameters_array': ( 'models.html#forger99model.get_parameters_array',
                                                                                           'circadian/models.py'),
                                  'circadian.models.Forger99Model.phase': ('models.html#forger99model.phase', 'circadian/models.py'),
                                  'circadian.models.Forger99Model.set_parameters': ( 'models.html#forger99model.set_parameters',
                                                                                     'circadian/models.py'),
                                  'circadian.models.Hannay19': ('models.html#hannay19', 'circadian/models.py'),
                                  'circadian.models.Hannay19.CBTObs': ('models.html#hannay19.cbtobs', 'circadian/models.py'),
                                  'circadian.models.Hannay19.DLMOObs': ('models.html#hannay19.dlmoobs', 'circadian/models.py'),
                                  'circadian.models.Hannay19.__init__': ('models.html#hannay19.__init__', 'circadian/models.py'),
                                  'circadian.models.Hannay19._default_params': ( 'models.html#hannay19._default_params',
                                                                                 'circadian/models.py'),
                                  'circadian.models.Hannay19.alpha0': ('models.html#hannay19.alpha0', 'circadian/models.py'),
                                  'circadian.models.Hannay19.amplitude': ('models.html#hannay19.amplitude', 'circadian/models.py'),
                                  'circadian.models.Hannay19.cbt': ('models.html#hannay19.cbt', 'circadian/models.py'),
                                  'circadian.models.Hannay19.default_initial_conditions': ( 'models.html#hannay19.default_initial_conditions',
                                                                                            'circadian/models.py'),
                                  'circadian.models.Hannay19.derv': ('models.html#hannay19.derv', 'circadian/models.py'),
                                  'circadian.models.Hannay19.dlmos': ('models.html#hannay19.dlmos', 'circadian/models.py'),
                                  'circadian.models.Hannay19.get_parameters': ( 'models.html#hannay19.get_parameters',
                                                                                'circadian/models.py'),
                                  'circadian.models.Hannay19.get_parameters_array': ( 'models.html#hannay19.get_parameters_array',
                                                                                      'circadian/models.py'),
                                  'circadian.models.Hannay19.integrate_observer': ( 'models.html#hannay19.integrate_observer',
                                                                                    'circadian/models.py'),
                                  'circadian.models.Hannay19.observer': ('models.html#hannay19.observer', 'circadian/models.py'),
                                  'circadian.models.Hannay19.phase': ('models.html#hannay19.phase', 'circadian/models.py'),
                                  'circadian.models.Hannay19.set_parameters': ( 'models.html#hannay19.set_parameters',
                                                                                'circadian/models.py'),
                                  'circadian.models.Hannay19TP': ('models.html#hannay19tp', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.CBTObs': ('models.html#hannay19tp.cbtobs', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.DLMOObs': ('models.html#hannay19tp.dlmoobs', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.__init__': ('models.html#hannay19tp.__init__', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP._default_params': ( 'models.html#hannay19tp._default_params',
                                                                                   'circadian/models.py'),
                                  'circadian.models.Hannay19TP.alpha0': ('models.html#hannay19tp.alpha0', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.amplitude': ('models.html#hannay19tp.amplitude', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.cbt': ('models.html#hannay19tp.cbt', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.default_initial_conditions': ( 'models.html#hannay19tp.default_initial_conditions',
                                                                                              'circadian/models.py'),
                                  'circadian.models.Hannay19TP.derv': ('models.html#hannay19tp.derv', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.dlmos': ('models.html#hannay19tp.dlmos', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.get_parameters': ( 'models.html#hannay19tp.get_parameters',
                                                                                  'circadian/models.py'),
                                  'circadian.models.Hannay19TP.get_parameters_array': ( 'models.html#hannay19tp.get_parameters_array',
                                                                                        'circadian/models.py'),
                                  'circadian.models.Hannay19TP.observer': ('models.html#hannay19tp.observer', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.phase': ('models.html#hannay19tp.phase', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.phase_difference': ( 'models.html#hannay19tp.phase_difference',
                                                                                    'circadian/models.py'),
                                  'circadian.models.Hannay19TP.set_parameters': ( 'models.html#hannay19tp.set_parameters',
                                                                                  'circadian/models.py')},
            'circadian.plots': { 'circadian.plots.Actogram': ('plots.html#actogram', 'circadian/plots.py'),
                                 'circadian.plots.Actogram.__init__': ('plots.html#actogram.__init__', 'circadian/plots.py'),
                                 'circadian.plots.Actogram.addLightSchedule': ( 'plots.html#actogram.addlightschedule',
                                                                                'circadian/plots.py'),
                                 'circadian.plots.Actogram.addRect': ('plots.html#actogram.addrect', 'circadian/plots.py'),
                                 'circadian.plots.Actogram.getRectangles': ('plots.html#actogram.getrectangles', 'circadian/plots.py'),
                                 'circadian.plots.Actogram.plot_phasemarker': ( 'plots.html#actogram.plot_phasemarker',
                                                                                'circadian/plots.py'),
                                 'circadian.plots.Actogram.plot_phasetimes': ('plots.html#actogram.plot_phasetimes', 'circadian/plots.py'),
                                 'circadian.plots.Stroboscopic': ('plots.html#stroboscopic', 'circadian/plots.py'),
                                 'circadian.plots.Stroboscopic.__init__': ('plots.html#stroboscopic.__init__', 'circadian/plots.py'),
                                 'circadian.plots.Stroboscopic._make_strobo_plot': ( 'plots.html#stroboscopic._make_strobo_plot',
                                                                                     'circadian/plots.py'),
                                 'circadian.plots.Stroboscopic.add_strobo_plot': ( 'plots.html#stroboscopic.add_strobo_plot',
                                                                                   'circadian/plots.py'),
                                 'circadian.plots.plot_mae': ('plots.html#plot_mae', 'circadian/plots.py'),
                                 'circadian.plots.plot_torus': ('plots.html#plot_torus', 'circadian/plots.py')},
            'circadian.readers': { 'circadian.readers.WearableData': ('readers.html#wearabledata', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.__getitem__': ( 'readers.html#wearabledata.__getitem__',
                                                                                   'circadian/readers.py'),
                                   'circadian.readers.WearableData.__post_init__': ( 'readers.html#wearabledata.__post_init__',
                                                                                     'circadian/readers.py'),
                                   'circadian.readers.WearableData._copy_with_metadata': ( 'readers.html#wearabledata._copy_with_metadata',
                                                                                           'circadian/readers.py'),
                                   'circadian.readers.WearableData.activity': ( 'readers.html#wearabledata.activity',
                                                                                'circadian/readers.py'),
                                   'circadian.readers.WearableData.aggregate': ( 'readers.html#wearabledata.aggregate',
                                                                                 'circadian/readers.py'),
                                   'circadian.readers.WearableData.build_sleep_chunks': ( 'readers.html#wearabledata.build_sleep_chunks',
                                                                                          'circadian/readers.py'),
                                   'circadian.readers.WearableData.date_bounds': ( 'readers.html#wearabledata.date_bounds',
                                                                                   'circadian/readers.py'),
                                   'circadian.readers.WearableData.datetime': ( 'readers.html#wearabledata.datetime',
                                                                                'circadian/readers.py'),
                                   'circadian.readers.WearableData.fillna': ('readers.html#wearabledata.fillna', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.filter': ('readers.html#wearabledata.filter', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.from_json': ( 'readers.html#wearabledata.from_json',
                                                                                 'circadian/readers.py'),
                                   'circadian.readers.WearableData.get_date': ( 'readers.html#wearabledata.get_date',
                                                                                'circadian/readers.py'),
                                   'circadian.readers.WearableData.get_timestamp': ( 'readers.html#wearabledata.get_timestamp',
                                                                                     'circadian/readers.py'),
                                   'circadian.readers.WearableData.groupby': ('readers.html#wearabledata.groupby', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.head': ('readers.html#wearabledata.head', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.heartrate': ( 'readers.html#wearabledata.heartrate',
                                                                                 'circadian/readers.py'),
                                   'circadian.readers.WearableData.join': ('readers.html#wearabledata.join', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.light_estimate': ( 'readers.html#wearabledata.light_estimate',
                                                                                      'circadian/readers.py'),
                                   'circadian.readers.WearableData.plot_heartrate': ( 'readers.html#wearabledata.plot_heartrate',
                                                                                      'circadian/readers.py'),
                                   'circadian.readers.WearableData.plot_hr_steps': ( 'readers.html#wearabledata.plot_hr_steps',
                                                                                     'circadian/readers.py'),
                                   'circadian.readers.WearableData.plot_light_activity': ( 'readers.html#wearabledata.plot_light_activity',
                                                                                           'circadian/readers.py'),
                                   'circadian.readers.WearableData.scatter_hr_steps': ( 'readers.html#wearabledata.scatter_hr_steps',
                                                                                        'circadian/readers.py'),
                                   'circadian.readers.WearableData.steps': ('readers.html#wearabledata.steps', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.steps_hr_loglinear': ( 'readers.html#wearabledata.steps_hr_loglinear',
                                                                                          'circadian/readers.py'),
                                   'circadian.readers.WearableData.tail': ('readers.html#wearabledata.tail', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.time_hour_bounds': ( 'readers.html#wearabledata.time_hour_bounds',
                                                                                        'circadian/readers.py'),
                                   'circadian.readers.WearableData.time_total': ( 'readers.html#wearabledata.time_total',
                                                                                  'circadian/readers.py'),
                                   'circadian.readers.WearableData.timestamp': ( 'readers.html#wearabledata.timestamp',
                                                                                 'circadian/readers.py'),
                                   'circadian.readers.WearableData.to_json': ('readers.html#wearabledata.to_json', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.trim_by_hour': ( 'readers.html#wearabledata.trim_by_hour',
                                                                                    'circadian/readers.py'),
                                   'circadian.readers.WearableData.trim_by_idx': ( 'readers.html#wearabledata.trim_by_idx',
                                                                                   'circadian/readers.py'),
                                   'circadian.readers.WearableData.trim_by_timestamp': ( 'readers.html#wearabledata.trim_by_timestamp',
                                                                                         'circadian/readers.py'),
                                   'circadian.readers.WearableData.utc_to_hrs': ( 'readers.html#wearabledata.utc_to_hrs',
                                                                                  'circadian/readers.py'),
                                   'circadian.readers.WearableData.wake': ('readers.html#wearabledata.wake', 'circadian/readers.py'),
                                   'circadian.readers.combine_wearable_streams': ( 'readers.html#combine_wearable_streams',
                                                                                   'circadian/readers.py'),
                                   'circadian.readers.read_actiwatch': ('readers.html#read_actiwatch', 'circadian/readers.py'),
                                   'circadian.readers.read_standard_csv': ('readers.html#read_standard_csv', 'circadian/readers.py'),
                                   'circadian.readers.read_standard_json': ('readers.html#read_standard_json', 'circadian/readers.py')},
            'circadian.sleep': { 'circadian.sleep.TwoProcessModel': ('sleep.html#twoprocessmodel', 'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.__call__': ('sleep.html#twoprocessmodel.__call__', 'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.__init__': ('sleep.html#twoprocessmodel.__init__', 'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.check_wake_status': ( 'sleep.html#twoprocessmodel.check_wake_status',
                                                                                        'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.dhomeostat': ( 'sleep.html#twoprocessmodel.dhomeostat',
                                                                                 'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.step_rk4': ('sleep.html#twoprocessmodel.step_rk4', 'circadian/sleep.py'),
                                 'circadian.sleep.cluster_sleep_periods_scipy': ( 'sleep.html#cluster_sleep_periods_scipy',
                                                                                  'circadian/sleep.py'),
                                 'circadian.sleep.sleep_midpoint': ('sleep.html#sleep_midpoint', 'circadian/sleep.py')},
            'circadian.utils': { 'circadian.utils.NpEncoder': ('utils.html#npencoder', 'circadian/utils.py'),
                                 'circadian.utils.NpEncoder.default': ('utils.html#npencoder.default', 'circadian/utils.py'),
                                 'circadian.utils.abs_hour_diff': ('utils.html#abs_hour_diff', 'circadian/utils.py'),
                                 'circadian.utils.angle_difference': ('utils.html#angle_difference', 'circadian/utils.py'),
                                 'circadian.utils.cal_days_diff': ('utils.html#cal_days_diff', 'circadian/utils.py'),
                                 'circadian.utils.circular_av_clock': ('utils.html#circular_av_clock', 'circadian/utils.py'),
                                 'circadian.utils.circular_mean': ('utils.html#circular_mean', 'circadian/utils.py'),
                                 'circadian.utils.circular_scatter': ('utils.html#circular_scatter', 'circadian/utils.py'),
                                 'circadian.utils.convert_binary': ('utils.html#convert_binary', 'circadian/utils.py'),
                                 'circadian.utils.cut_phases_12': ('utils.html#cut_phases_12', 'circadian/utils.py'),
                                 'circadian.utils.interpolateLinear': ('utils.html#interpolatelinear', 'circadian/utils.py'),
                                 'circadian.utils.interpolateLinearExt': ('utils.html#interpolatelinearext', 'circadian/utils.py'),
                                 'circadian.utils.parse_dt': ('utils.html#parse_dt', 'circadian/utils.py'),
                                 'circadian.utils.phase_coherence': ('utils.html#phase_coherence', 'circadian/utils.py'),
                                 'circadian.utils.phase_coherence_clock': ('utils.html#phase_coherence_clock', 'circadian/utils.py'),
                                 'circadian.utils.phase_ic_guess': ('utils.html#phase_ic_guess', 'circadian/utils.py'),
                                 'circadian.utils.simple_norm_stepshr_sleep_classifier': ( 'utils.html#simple_norm_stepshr_sleep_classifier',
                                                                                           'circadian/utils.py'),
                                 'circadian.utils.split_drop_data': ('utils.html#split_drop_data', 'circadian/utils.py'),
                                 'circadian.utils.split_missing_data': ('utils.html#split_missing_data', 'circadian/utils.py'),
                                 'circadian.utils.subtract_clock_times': ('utils.html#subtract_clock_times', 'circadian/utils.py'),
                                 'circadian.utils.times_to_angle': ('utils.html#times_to_angle', 'circadian/utils.py'),
                                 'circadian.utils.timezone_mapper': ('utils.html#timezone_mapper', 'circadian/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_metrics.ipynb.

# %% auto 0
__all__ = ['esri', 'esri_trajectory']

# %% ../nbs/04_metrics.ipynb 3
import numpy as np
from .models import Hannay19
from .utils import phase_ic_guess
from .readers import WearableData

# %% ../nbs/04_metrics.ipynb 4
def esri(awObj: WearableData, 
                gamma: float = 0.0,
                multiplier: float = 1.0,
                num_days: float = 4.5):

    spmodel = Hannay19({'K': 0.0, 'gamma': gamma})
    psi0 = phase_ic_guess(awObj.time_total[0])

    idx = awObj.time_total < awObj.time_total[0]+24*num_days
    sol = spmodel.integrate_model(
        awObj.time_total[idx], multiplier*awObj.steps[idx],
        np.array([0.10, psi0, 0.0])
    )
    return sol.ts, sol.states



# %% ../nbs/04_metrics.ipynb 5
def esri_trajectory(awObj: WearableData,
                           gamma: float = 0.0,
                           multiplier: float = 1.0,
                           num_days: float = 4.5,
                           ):

    spmodel = Hannay19({'K': 0.0, 'gamma': gamma})
    compactness_trajectory = []
    time_trajectory = []
    timestamps = []
    timeStart = awObj.time_total[0]
    timestampStart = awObj.timestamp[0]
    while timeStart < awObj.time_total[-1] - 24*num_days:
        try:
            psi0 = phase_ic_guess(timeStart)
            idxStart = (awObj.time_total > timeStart)
            tsFilter = awObj.time_total[idxStart]
            idx = tsFilter < np.array(tsFilter[0])+24*num_days
            stepsFilter = awObj.steps[idxStart]
            trajectory = spmodel.integrate_model(
                tsFilter[idx], multiplier*stepsFilter[idx],
                np.array([0.10, psi0, 0.0])
            )
            sol = trajectory.states
            if sol[0, -1] > 0.0:
                compactness_trajectory.append(sol[0, -1])
                time_trajectory.append(timeStart)
                timestamps.append(
                    3600.0*(timeStart - awObj.time_total[0]) + timestampStart)
        except:
            print("Error in trajectory")
        timeStart += 1.0
    return time_trajectory, timestamps, compactness_trajectory




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_lights.ipynb.

# %% auto 0
__all__ = ['Light', 'make_pulse', 'get_pulse', 'RimmerLightPulseLight', 'PhaseResponseCurveLight', 'heaviside', 'PRCFinder',
           'IntensityResponseCurveLight', 'DosageResponseCurve']

# %% ../nbs/01_lights.ipynb 3
from matplotlib.pyplot import step
import numpy as np
from numpy.core.fromnumeric import repeat
import pylab as plt
from typing import Optional

# %% ../nbs/01_lights.ipynb 4
class Light:

    def __init__(self, 
                 light: callable, # light function that takes in a time value and returns a float, if a float is passed, then the light function is a constant set to that lux value 
                 start_time: float = 0.0, # when the light function starts in hours
                 duration: float = 1.0, # duration of the light function in hours
                 default_value=0.0 # the default value of the light function when it is not active
                 ):
        if type(light) is float:
            self._func = np.vectorize(lambda t: light, otypes=[float])
        else:
            self._func = np.vectorize(light, otypes=[float])
        self.default_value = default_value
        self.start_time = start_time
        self.duration = duration

    @property
    def end_time(self):
        return self.start_time + self.duration

    def __call__(self, 
                 t: np.ndarray, # time values in hours to evaluate the light function
                 repeat_period: float = None # should the light function repeat after a certain period
                 ):
        t = np.array(t)
        if repeat_period is not None:
            t = np.mod(t, repeat_period)
        default_result = np.zeros_like(t) + self.default_value
        mask = (t >= self.start_time) & (t <= self.end_time)
        func_result  = self._func(t[mask])
        default_result[mask] += func_result
        return default_result

    def __add__(self, light_obj: 'Light'):
        switch_time = self.end_time
        def light_func_new(t): return np.piecewise(t, [t <= switch_time, t >= switch_time],                                       [
            self._func, lambda t: light_obj._func(t-switch_time)])

        start_time_new = min(self.start_time, light_obj.start_time)
        duration_new = light_obj.duration + self.duration
        return Light(light_func_new, start_time=start_time_new, duration=duration_new)

    def plot(self,
             start_time: float = None, 
             end_time: float = None,
             ax=None, 
             show_plot=True, 
             num_samples=10000,
             *args,
             **kwargs):
        
        if start_time is None:
            start_time = self.start_time
        if end_time is None:
            end_time = self.end_time
            
        t = np.linspace(start_time, end_time, num_samples)
        vals = self.__call__(t)
        if ax is None:
            plt.figure()
            ax = plt.gca()

        ax.plot(t, vals, *args, **kwargs)
        if show_plot:
            plt.show()
        return ax
    
    @staticmethod
    def RegularLight(lux=150.0, # lux intensity of the light
                 lights_on=8.0, # hour of the day for lights to come on
                 lights_off=16.0, # hour of the day for lights to go off
                 ) -> 'Light':

        if lights_off > lights_on:
            end_time = lights_on + (lights_off - lights_on) 
            dark_section_before = Light(0.0, start_time=0.0, duration=lights_on)
            light_section = Light(lux, 
                         start_time=lights_on, 
                         duration=lights_off - lights_on)
            dark_section = Light(0.0, end_time, duration=24.0 - end_time)
            return dark_section_before + light_section + dark_section 
        else:
            first_light = Light(lux, start_time = 0.0, duration=lights_off)
            dark_section = Light(0.0, start_time=lights_off, duration=lights_on - lights_off)
            second_light = Light(lux, start_time=lights_on, duration=24.0 - lights_on)
            return  first_light + dark_section + second_light
        
    @staticmethod
    def ShiftWorkLight(lux: float = 150.0, # lux intensity of the light
                   dayson: int = 3, # number of days on the night shift
                   daysoff: int = 2 # number of days off shift
                   ) -> 'Light':
        
        workday = Light.RegularLight(lux=lux, lights_on=19.0, lights_off=11.0)
        offday = Light.RegularLight(lux=lux, lights_on=7.0, lights_off=23.0)
        total_schedule = [workday for _ in range(dayson-1)] + [offday for _ in range(daysoff)]
        for day in total_schedule:
            workday += day
        return workday
    
    @staticmethod
    def SlamShift(lux: float = 150.0, # lux intensity of the light
                  shift=8.0, # number of hours to shift the light schedule
                  before_days=10, #number of days before the shift occurs 
                  starting_lights_on=8.0, # hour of the day for lights to come on
                  after_days=10 # number of days after the shift occurs
                  ) -> 'Light':
        
        light_before = Light.RegularLight(lux=lux, 
                                          lights_on=starting_lights_on, 
                                          lights_off= np.fmod(starting_lights_on + 16.0, 24.0)
                                          )
        light_after = Light.RegularLight(lux=lux, 
                                         lights_on=np.fmod(starting_lights_on+shift, 24.0), 
                                         lights_off=np.fmod(starting_lights_on+shift+16.0, 24.0)
                                         )
        total_schedule = [light_before for _ in range(before_days-1)] + [light_after for _ in range(after_days)]
        for day in total_schedule:
            light_before += day
        return light_before
    



# %% ../nbs/01_lights.ipynb 7
def make_pulse(t, tstart, tend, steep: float = 30.0):
    return 0.5*np.tanh(steep*(t-tstart))-0.5*np.tanh(steep*(t-tend))

def get_pulse(t: float,
              t1: float,
              t2: float,
              repeat=False,
              Intensity: float = 150.0):

    if repeat:
        t = np.fmod(t, 24.0)
    if t < 0.0:
        t += 24.0

    light_value = Intensity*make_pulse(t, t1, t2)
    return np.abs(light_value)


# %% ../nbs/01_lights.ipynb 18
# Need to finish implementing this 


class RimmerLightPulseLight:
     
    def __init__(self) -> None:
        pass 
      
    @staticmethod
    def make_pulse_rimmer(t: float, 
                          tstart: float, 
                          tend: float):
        val = 0.0
        steep = 100.0
        trans = 25.0/60.0

        val += 9500*(0.5*np.tanh(steep*(t-tstart))-0.5*np.tanh(steep*(t-t)))

        if ((t < tstart) and (t > tstart-trans)):
            val = 5000.0
            factor = np.floor((t-tstart+trans)*60.0/5.0)
            val += factor*1000.0

        if ((t > t) and (t < t+trans)):
            val = 9500.0
            factor = np.floor((t-t)*60.0/5.0)
            val -= factor*1000.0

        return(val)

    @staticmethod
    def pulse_rimmer_start(t: float, 
                             tstart: float, 
                             tend: float):

        val = 0.0
        steep = 100.0
        trans = 25.0/60.0

        val += 9500*(0.5*np.tanh(steep*(t-tstart))-0.5 *
                    np.tanh(steep*(t-t)))  # main pulse

        if ((t < tstart) and (t > tstart-trans)):
            val = 5000.0
            factor = np.floor((t-tstart+trans)*60.0/5.0)
            val += factor*1000.0

        return(val)




# %% ../nbs/01_lights.ipynb 19
class PhaseResponseCurveLight:
    
    def __init__(self) -> None:
         pass 
     
    def light_khalsa(t: float, 
                     CR: float):
        # Implement a Khalsa Light Schedule give a parameter for the initial CR length */

        low_light = 15.0  # from the paper
        val = low_light*make_pulse(t, 0.0, CR)+low_light*make_pulse(t,
                                                                    CR+8.0, CR+24.0)+low_light*make_pulse(t, CR+32.0, CR+1000.0)
        val += 9985.0*make_pulse(t, CR+8.0+4.65, CR+8.0+11.35)
        return(val)

    def light_hilaire(t: float,
                      CR: float):

        low_light = 3.0  # amount of light during CR (verified 1/16/17)
        val = low_light*make_pulse(t, 0, CR)+low_light*make_pulse(t,
                                                                CR+8, CR+24)+low_light*make_pulse(t, CR+32, CR+1000)
        val += 7997*make_pulse(t, CR+8+7.5, CR+8+8.5)
        return(val)

    def light_amplitude_reseting(t: float, 
                  CR: float):

        roomLight = 150.0
        afterLight = 150.0
        beforeLight = 150.0
        val = 0.0

        val += beforeLight*make_pulse(t, 0, CR)
        val += roomLight*make_pulse(t, CR+8, CR+24)
        val += 0.02*make_pulse(t, 0+24.0, CR+24.0)+roomLight * \
            make_pulse(t, CR+8+24.0, CR+48.0)
        val += afterLight*make_pulse(t, CR+48.0+8.0, CR+24+48.0+1000)
        val += 9850*make_pulse(t, CR+8+8-2.5, CR+8+8+2.5) + \
            9850*make_pulse(t, CR+8+8-2.5+24.0, CR+8+8+2.5+24.0)
        return(val)
    
    def light_czeiler_type0(t: float, CR: float):
        roomLight = 150.0
        afterLight = 150.0  # 150.0
        beforeLight = 150.0  # 150.0

        val = 0.0

        val += beforeLight*make_pulse(t, 0, CR)
        val += roomLight*make_pulse(t, CR+8, CR+24)
        val += 0.02*make_pulse(t, 0+24.0, CR+24.0)+roomLight * \
            make_pulse(t, CR+8+24.0, CR+24+24.0)
        val += 0.02*make_pulse(t, 0+48.0, CR+48.0)+roomLight * \
            make_pulse(t, CR+8+48.0, CR+24+48.0)
        val += afterLight*make_pulse(t, CR+24+48.0+8.0, CR+24+48.0+1000)
        val += 9850*make_pulse(t, CR+8+8-2.5, CR+8+8+2.5)+9850*make_pulse(t, CR+8+8 -
                                                                        2.5+24.0, CR+8+8+2.5+24.0)+9850*make_pulse(t, CR+8+8-2.5+48.0, CR+8+8+2.5+48.0)
        return(val)
    
    
    
    



# %% ../nbs/01_lights.ipynb 20
from .models import CircadianModel


def heaviside(x: float) -> float:
    if x < 0:
        return 0
    else:
        return 1
    
class PRCFinder:
    
    @property
    def type0x(self) -> np.ndarray:
        return [16.2593, 16.304, 16.3018, 16.2738, 16.493, 16.7512, 17.9321, 19.1632,
                20.0534, 20.2959, 20.4188, 20.6394, 20.787, 21.1123, 21.1377,
                21.1579, 21.183, 21.2433, 21.2689, 22.2468, 25.2147, 27.1642,
                27.1974, 27.2108, 27.2363, 27.257, 27.5736, 27.8422, 28.0516,
                28.3226, 28.3779, 28.4506, 28.6003, 28.6608, 28.7964, 28.8359,
                28.8422, 29.0239, 29.0813, 29.1065, 29.1506, 29.2144, 29.2561,
                29.4005, 29.8354, 26.0773, 25.5784, 24.256, 20.0592, 15.2187,
                6.93359, 6.95777, 6.98051, 7.00128, 7.08384, 7.99415, 8.25918,
                8.7184, 8.91649, 8.97987, 9.2621, 9.30382, 9.47454, 10.3813, 10.8293,
                10.8903, 10.9258, 11.0337, 11.8089, 16.15, 16.3256, 16.363, 16.3653,
                16.6808, 16.7064]
    
    def exp_type0(x: float, b: float):
        val=(-1.57154+0.228932/(-1.0*b+x)-0.650632*x)*heaviside(b-x)+(9.66876+0.1321196/(-1.0*b+x)-0.463105*x)*heaviside(x-b)
        if (val < -12.0):
            return(-12.0)
        if (val > 12.0):
            return(12.0)

        return val
    
    @staticmethod
    def prc_type0_point(CRlength: float, initial_value: np.ndarray, model: CircadianModel) -> float:
        """
        Gives a single point from the Type 0 Human protocol. The initial cond
        passed in should already be entrained to a regular routine.
        CRlengths for the model fit were taken to vary in the range (6.0,30.0)
        every 0.05. However, this was just for convience.
        Every data point in this study is actually a diffirent person so the
        parameters should be allowed to vary for each computed point.....
        """
        
        CRFinal=30.0
        tend=72.0+8.0+30.0+CRFinal
        ts = np.arange(0,tend,0.10)
        light_vals = np.array([PhaseResponseCurveLight.light_czeiler_type0(t, CRlength) for t in ts])
        trajectory = model(ts, light_vals, initial_value)
        CBT = model.cbt(trajectory=trajectory)
        shift=(CBT[0]-CBT[-1]) % 24.0 #finds a neg number between zero and -24.0

        #now shifts live in the range (-12,12) hours
        if (shift < -12.0):
            shift+=24.0

        #Now find the phase of the stimulus
        phase=(CRlength-CBT[0]+12.0+16.0-2.5) % 24.0
        return phase,shift
    
    
        

# %% ../nbs/01_lights.ipynb 22
class IntensityResponseCurveLight:
    
    def __init__(self) -> None:
         pass
     
    @staticmethod
    def light_intensity(t, Intensity):
        """
        Define the light schedule for the dosage response curve experiments. All light exposures started at phi 6.75 hours
        before Tmin and lasted 6.5 hours of varying intensities
        """

        cr_light_level = 10.0  # verified, in the statistics section of the methods
        wake_light_level = 10.0  # verified
        wake_stimulus_light_level = 0.03
        sleep_light_level = 0.03
        stimulus_light_level = Intensity

        w = 50.0  # length of the constant routine, should be approx 50 hours, did have 50.0-9.50
        val = 0.0
        val += cr_light_level*make_pulse(t, 0, w)  # 50 hour constant routine
        val += sleep_light_level*make_pulse(t, w, w+8.0)  # eight hour sleep bout
        val += wake_stimulus_light_level * \
            make_pulse(t, w+8.0, w+24.0)  # wake/stimulus period
        val += (stimulus_light_level-wake_stimulus_light_level)*make_pulse(t, w+16.0 -
                                                                        3.25, w+16.0+3.25)  # the stimulus centered during the 16 hours of wakefulness
        # sleep period following stimulus
        val += sleep_light_level*make_pulse(t, w+24.0, w+32.0)
        # final cr to assess the phase shift induced
        val += cr_light_level*make_pulse(t, w+32.0, w+32.0+30.0)

        # Stimulus should start 5.25 hours after phase zero, so that it is centered 3.5 hours before Tmin
        return(val)



# %% ../nbs/01_lights.ipynb 23
class DosageResponseCurve:
    
    def __init__(self) -> None:
         pass
     
    @staticmethod
    def light_dosage(t: float, length: float):
        # Implement the Light Schedule for Chang et al's Dosage Response Curve */

        CR1 = 48.0  # approx 50.0
        CR2 = 30.0
        CR_light_level = 1.0
        stimulus_light_level = 10000.0
        wake_light_level = 3.0

        val = 0.0
        val += CR_light_level*make_pulse(t, 0.0, CR1)
        val += wake_light_level*make_pulse(t, CR1+8.0, CR1+8.0+16.0)
        if (length == 0.2):
            val += stimulus_light_level * \
                make_pulse(t, CR1+8.0+8.5-length/2.0-12.0/60.0,
                            CR1+8.0+8.5+length/2.0-12.0/60.0)
        else:
            val += stimulus_light_level * \
                make_pulse(t, CR1+8.0+8.5-length/2.0, CR1+8.0+8.5+length/2.0)

        val += CR_light_level*make_pulse(t, CR1+32.0, CR1+32.0+30.0)

        return(val)

    @staticmethod
    def light_dosage_day1(t):
        # Implement the prep day just before the constant routine in the Chang et al Duration Response Curve protocol */
        s = t % 24.0
        val = 0.0
        if (s < 8.0):
            val = 90.0
        if ((s >= 8.0) and (s <= 16.0)):
            val = 3.0
        return(val)


